import os
import shutil
from pathlib import Path

from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.core.tasks import BaseTask
from cumulusci.utils.options import CCIOptions, Field


class DirectoryRecreator(BaseTask):
    class Options(CCIOptions):
        path: Path = Field(..., description="Path to the directory to recreate")

    parsed_options: Options

    def _init_options(self, kwargs):
        super()._init_options(kwargs)

        if os.path.isfile(self.parsed_options.path):
            raise TaskOptionsError(f"Path {self.parsed_options.path} is a file")

    def _run_task(self):
        created = "created"
        if os.path.exists(self.parsed_options.path):
            shutil.rmtree(self.parsed_options.path)
            created = "removed and created"

        os.makedirs(self.parsed_options.path, exist_ok=True)

        self.logger.info(f"Directory {self.parsed_options.path} {created}.")



import base64
import json
import os
import pathlib
import shutil
from unittest import mock

import pytest  # noqa: F401
from simple_salesforce.exceptions import SalesforceMalformedRequest

from cumulusci.core.exceptions import CumulusCIException  # noqa: F401
from cumulusci.tasks.salesforce.tests.util import create_task
from cumulusci.tasks.salesforce.users.photos import UploadProfilePhoto, join_errors
from cumulusci.utils import temporary_dir


def test_join_errors():
    e = SalesforceMalformedRequest(
        "url",
        "status",
        "resource_name",
        [
            {"message": "Error message 1.", "errorCode": "ERROR_1"},
            {"message": "Error message 2.", "errorCode": "ERROR_2"},
            {"message": "Error message 3.", "errorCode": "ERROR_3"},
        ],
    )

    assert join_errors(e) == "Error message 1.; Error message 2.; Error message 3."


class TestUploadProfilePhoto:
    def setup_method(self):
        # Photo information
        self.photo = "photo.mock.txt"
        self.base_path = os.path.dirname(__file__)
        self.photo_path = pathlib.Path(
            os.path.join(self.base_path, self.photo)
        ).resolve()
        assert self.photo_path.exists(), "photo mock cannot be found"
        assert self.photo_path.is_file(), "photo_path should point to a file"

        # Resusable data
        self.content_version_id = "0681k000001YWQ5AAO"
        self.content_document_id = "0691k000001LtfEAAS"

        self.e = SalesforceMalformedRequest(
            "url",
            "status",
            "resource_name",
            [
                {"message": "Error message 1.", "errorCode": "ERROR_1"},
                {"message": "Error message 2.", "errorCode": "ERROR_2"},
                {"message": "Error message 3.", "errorCode": "ERROR_3"},
            ],
        )
        self.expected_exception_message = (
            "Error message 1.; Error message 2.; Error message 3."
        )
        self.user_id = "0051k000003cEL3AAM"

        # Task with mocks
        self.task = create_task(UploadProfilePhoto, {"photo": self.photo})
        self.task.logger = mock.Mock()
        self.task.sf = mock.Mock()

    def test_raise_cumulusci_exception(self):
        with pytest.raises(CumulusCIException) as e:
            self.task._raise_cumulusci_exception(self.e)

        assert e.value.args[0] == self.expected_exception_message

    def test_get_user_id_by_query__valid_query__0_records(self):
        # where is prefixed with "where " which will be stripped when injected into the query.
        where = "where (FirstName = 'User' OR LastName = 'User') AND IsActive = true AND CreatedDate = TODAY"
        expected_query = "SELECT Id FROM User WHERE (FirstName = 'User' OR LastName = 'User') AND IsActive = true AND CreatedDate = TODAY LIMIT 2"

        self.task.sf.query_all.return_value = {"records": []}

        self.task.logger.info._expected_calls = [
            mock.call(f"Querying User: {expected_query}")
        ]

        # Execute the test
        with pytest.raises(CumulusCIException) as e:
            self.task._get_user_id_by_query(where)

        assert e.value.args[0] == "No Users found."

        self.task.sf.query_all.assert_called_once_with(expected_query)
        self.task.logger.info.assert_has_calls(self.task.logger.info._expected_calls)

    def test_get_user_id_by_query__valid_query__1_record(self):
        # where is prefixed with "where " which will be stripped when injected into the query.
        where = "where (FirstName = 'User' OR LastName = 'User') AND IsActive = true AND CreatedDate = TODAY"
        expected_query = "SELECT Id FROM User WHERE (FirstName = 'User' OR LastName = 'User') AND IsActive = true AND CreatedDate = TODAY LIMIT 2"

        expected = self.user_id

        self.task.sf.query_all.return_value = {"records": [{"Id": expected}]}

        self.task.logger.info._expected_calls = [
            mock.call(f"Querying User: {expected_query}"),
            mock.call(f"Uploading profile photo for the User with ID {expected}"),
        ]

        # Execute the test
        actual = self.task._get_user_id_by_query(where)

        assert expected == actual

        self.task.sf.query_all.assert_called_once_with(expected_query)
        self.task.logger.info.assert_has_calls(self.task.logger.info._expected_calls)

    def test_get_user_id_by_query__valid_query__2_records(self):
        # where is prefixed with "where " which will be stripped when injected into the query.
        where = "where (FirstName = 'User' OR LastName = 'User') AND IsActive = true AND CreatedDate = TODAY"
        expected_query = "SELECT Id FROM User WHERE (FirstName = 'User' OR LastName = 'User') AND IsActive = true AND CreatedDate = TODAY LIMIT 2"

        id_0 = self.user_id
        id_1 = "0051k000003cEL8AAM"

        self.task.sf.query_all.return_value = {"records": [{"Id": id_0}, {"Id": id_1}]}

        self.task.logger.info._expected_calls = [
            mock.call(f"Querying User: {expected_query}")
        ]

        # Execute the test
        with pytest.raises(CumulusCIException) as e:
            self.task._get_user_id_by_query(where)

        assert (
            e.value.args[0] == f"More than one User found (at least 2): {id_0}, {id_1}"
        )

        self.task.sf.query_all.assert_called_once_with(expected_query)
        self.task.logger.info.assert_has_calls(self.task.logger.info._expected_calls)

    def test_get_user_id_by_query__invalid_query(self):
        where = "(FirstName = 'User' OR LastName = 'User') AND IsActive = true AND CreatedDate = TODAY but the end of this query is invalid SOQL "
        expected_query = "SELECT Id FROM User WHERE (FirstName = 'User' OR LastName = 'User') AND IsActive = true AND CreatedDate = TODAY but the end of this query is invalid SOQL  LIMIT 2"

        self.task.sf.query_all.side_effect = self.e

        self.task.logger.info._expected_calls = [
            mock.call(f"Querying User: {expected_query}")
        ]

        # Execute the test
        with pytest.raises(CumulusCIException) as e:
            self.task._get_user_id_by_query(where)

        assert e.value.args[0] == self.expected_exception_message

        self.task.sf.query_all.assert_called_once_with(expected_query)
        self.task.logger.info.assert_has_calls(self.task.logger.info._expected_calls)

    def test_get_default_user_id(self):
        expected = self.user_id

        self.task.sf.restful.return_value = {
            "identity": f"a bunch of other things{expected}"
        }

        actual = self.task._get_default_user_id()

        assert expected == actual

        self.task.logger.info.assert_called_once_with(
            f"Uploading profile photo for the default User with ID {expected}"
        )

    def test_insert_content_document__no_photo_found(self):
        with temporary_dir():
            with pytest.raises(CumulusCIException) as e:
                self.task._insert_content_document(self.photo)

            assert e.value.args[0] == f"No photo found at {self.photo}"

    def test_insert_content_document__exception_creating_content_document(self):
        errors = "the reason why inserting the ContentVersion failed"
        self.task.sf.ContentVersion.create.return_value = {
            "success": False,
            "errors": errors,
        }

        with temporary_dir() as d:
            # Copy photo to temporary direcory.
            shutil.copy(str(self.photo_path), d)
            temp_photo_path = pathlib.Path(os.path.join(d, self.photo)).resolve()
            assert (
                temp_photo_path.exists()
            ), "photo mock was not copied to the temporary directory"
            assert (
                temp_photo_path.is_file()
            ), "Path the the photo mock in the temporary directory should point to a file"

            self.task.sf.ContentVersion.create._expected_calls = [
                mock.call(
                    {
                        "PathOnClient": temp_photo_path.name,
                        "Title": temp_photo_path.stem,
                        "VersionData": base64.b64encode(
                            temp_photo_path.read_bytes()
                        ).decode("utf-8"),
                    }
                )
            ]

            with pytest.raises(CumulusCIException) as e:
                self.task._insert_content_document(self.photo)

        assert e.value.args[0] == f"Failed to create photo ContentVersion: {errors}"

        self.task.sf.ContentVersion.create.assert_has_calls(
            self.task.sf.ContentVersion.create._expected_calls
        )

        self.task.logger.info.assert_called_once_with(
            f"Setting user photo to {self.photo}"
        )

    def test_insert_content_document__success(self):
        self.task.sf.ContentVersion.create.return_value = {
            "success": True,
            "id": self.content_version_id,
        }
        self.task.sf.query.return_value = {
            "records": [
                {
                    "Id": self.content_version_id,
                    "ContentDocumentId": self.content_document_id,
                }
            ]
        }
        self.task.logger.info._expected_calls = [
            mock.call(f"Setting user photo to {self.photo}"),
            mock.call(
                f"Uploaded profile photo ContentDocument {self.content_document_id}"
            ),
        ]

        expected = self.content_document_id

        with temporary_dir() as d:
            # Copy photo to temporary direcory.
            shutil.copy(str(self.photo_path), d)
            temp_photo_path = pathlib.Path(os.path.join(d, self.photo)).resolve()
            assert (
                temp_photo_path.exists()
            ), "photo mock was not copied to the temporary directory"
            assert (
                temp_photo_path.is_file()
            ), "Path the the photo mock in the temporary directory should point to a file"

            self.task.sf.ContentVersion.create._expected_calls = [
                mock.call(
                    {
                        "PathOnClient": temp_photo_path.name,
                        "Title": temp_photo_path.stem,
                        "VersionData": base64.b64encode(
                            temp_photo_path.read_bytes()
                        ).decode("utf-8"),
                    }
                )
            ]

            actual = self.task._insert_content_document(self.photo)

        assert expected == actual

        self.task.sf.ContentVersion.create.assert_has_calls(
            self.task.sf.ContentVersion.create._expected_calls
        )

        self.task.sf.query.assert_called_once_with(
            f"SELECT Id, ContentDocumentId FROM ContentVersion WHERE Id = '{self.content_version_id}'"
        )

        self.task.logger.info.assert_has_calls(self.task.logger.info._expected_calls)

    def test_delete_content_document(self):
        self.task._delete_content_document(self.content_document_id)

        self.task.sf.ContentDocument.delete.assert_called_once_with(
            self.content_document_id
        )

    def test_assign_user_profile_photo__exception(self):
        self.task.sf.restful.side_effect = self.e
        self.task.logger.error._expected_calls = [
            mock.call(
                "An error occured assigning the ContentDocument as the users's profile photo."
            ),
            mock.call(f"Deleting ContentDocument {self.content_document_id}"),
        ]
        self.task._delete_content_document = mock.Mock()

        with pytest.raises(CumulusCIException) as e:
            self.task._assign_user_profile_photo(self.user_id, self.content_document_id)

        assert e.value.args[0] == self.expected_exception_message

        self.task._delete_content_document.assert_called_once_with(
            self.content_document_id
        )

        self.task.logger.error.assert_has_calls(self.task.logger.error._expected_calls)

    def test_assign_user_profile_photo__success(self):
        self.task._delete_content_document = mock.Mock()

        self.task._assign_user_profile_photo(self.user_id, self.content_document_id)

        self.task.sf.restful.assert_called_once_with(
            f"connect/user-profiles/{self.user_id}/photo",
            data=json.dumps({"fileId": self.content_document_id}),
            method="POST",
        )

        self.task._delete_content_document.assert_not_called()

        self.task.logger.error.assert_not_called()

    def test_run_task__default_user(self):
        assert bool(self.task.options.get("where")) is False

        self.task._get_default_user_id = mock.Mock()
        self.task._get_user_id_by_query = mock.Mock()
        self.task._insert_content_document = mock.Mock()
        self.task._assign_user_profile_photo = mock.Mock()

        self.task._run_task()

        self.task._get_default_user_id.assert_called_once_with()

        self.task._get_user_id_by_query.assert_not_called()

        self.task._insert_content_document.assert_called_once_with(
            self.task.options["photo"]
        )

        self.task._assign_user_profile_photo.assert_called_once_with(
            self.task._get_default_user_id.return_value,
            self.task._insert_content_document.return_value,
        )

    def test_run_task__queries_user(self):
        self.task.options["where"] = mock.Mock()
        assert bool(self.task.options.get("where")) is True

        self.task._get_default_user_id = mock.Mock()
        self.task._get_user_id_by_query = mock.Mock()
        self.task._insert_content_document = mock.Mock()
        self.task._assign_user_profile_photo = mock.Mock()

        self.task._run_task()

        self.task._get_default_user_id.assert_not_called()

        self.task._get_user_id_by_query.assert_called_once_with(
            self.task.options["where"]
        )

        self.task._insert_content_document.assert_called_once_with(
            self.task.options["photo"]
        )

        self.task._assign_user_profile_photo.assert_called_once_with(
            self.task._get_user_id_by_query.return_value,
            self.task._insert_content_document.return_value,
        )



import io
import os
import pathlib
import zipfile
from base64 import b64encode
from tempfile import TemporaryDirectory
from unittest import mock

import pytest
import responses
import yaml

from cumulusci.core.config import BaseProjectConfig, ServiceConfig, UniversalConfig
from cumulusci.core.dependencies.github_resolvers import VCS_GITHUB
from cumulusci.core.exceptions import DependencyResolutionError, VcsApiError
from cumulusci.core.keychain import BaseProjectKeychain
from cumulusci.tasks.release_notes.tests.utils import MockUtilBase
from cumulusci.utils import temporary_dir, touch
from cumulusci.utils.yaml.cumulusci_yml import (
    GitHubSourceModel,
    GitHubSourceRelease,
    LocalFolderSourceModel,
)

from ..source import GitHubEnterpriseSource, GitHubSource, LocalFolderSource


class TestGitHubSource(MockUtilBase):
    def setup_method(self):
        self.repo_api_url = "https://api.github.com/repos/TestOwner/TestRepo"
        universal_config = UniversalConfig()
        self.project_config = BaseProjectConfig(
            universal_config, repo_info={"root": os.getcwd()}
        )
        self.project_config.set_keychain(BaseProjectKeychain(self.project_config, None))
        self.project_config.config["sources"]["foo"] = {
            "github": "https://github.com/TestOwner/Foo",
            "release": "latest",
        }
        self.repo_root = TemporaryDirectory()
        self.project_config.keychain.set_service(
            "github",
            "test_alias",
            ServiceConfig(
                {
                    "username": "TestUser",
                    "token": "TestPass",
                    "email": "testuser@testdomain.com",
                }
            ),
        )

    def teardown_method(self):
        self.repo_root.cleanup()

    @responses.activate
    def test_resolve__default(self):
        # The default is to use resolution strategy `production`, which gets the latest release.
        responses.add(
            method=responses.GET,
            url=self.repo_api_url,
            json=self._get_expected_repo(owner="TestOwner", name="TestRepo"),
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/releases/latest",
            json=self._get_expected_release("release/1.0"),
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/release/1.0",
            json=self._get_expected_tag_ref("release/1.0", "tag_sha"),
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/tags/tag_sha",
            json=self._get_expected_tag("release/1.0", "commit_sha", "tag_sha"),
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/contents/cumulusci.yml?ref=commit_sha",
            json={
                "url": "https://api.github.com/repos/TestOwner/TestRepo/contents/cumulusci.yml?ref=commit_sha",
                "download_url": "https://api.github.com/repos/TestOwner/TestRepo/contents/cumulusci.yml?ref=commit_sha",
                "git_url": "https://api.github.com/repos/TestOwner/TestRepo/contents/cumulusci.yml?ref=commit_sha",
                "html_url": "https://api.github.com/repos/TestOwner/TestRepo/contents/cumulusci.yml?ref=commit_sha",
                "_links": {},
                "name": "cumulusci.yml",
                "path": "cumulusci.yml",
                "sha": "commit_sha",
                "size": 100,
                "type": "yaml",
                "encoding": "base64",
                "content": b64encode(
                    yaml.dump(
                        {
                            "project": {
                                "package": {
                                    "name_managed": "Test Product",
                                    "namespace": "ns",
                                }
                            }
                        }
                    ).encode("utf-8")
                ).decode("utf-8"),
            },
        )

        source = GitHubSource(
            self.project_config,
            GitHubSourceModel(github="https://github.com/TestOwner/TestRepo.git"),
        )
        assert source.vcs == VCS_GITHUB
        assert repr(source) == "<GitHubSource GitHub: TestOwner/TestRepo @ commit_sha>"

    @responses.activate
    def test_resolve__commit(self):
        responses.add(
            method=responses.GET,
            url=self.repo_api_url,
            json=self._get_expected_repo(owner="TestOwner", name="TestRepo"),
        )

        source = GitHubSource(
            self.project_config,
            GitHubSourceModel(
                github="https://github.com/TestOwner/TestRepo.git", commit="abcdef"
            ),
        )
        assert repr(source) == "<GitHubSource GitHub: TestOwner/TestRepo @ abcdef>"

    @responses.activate
    def test_resolve__ref(self):
        responses.add(
            method=responses.GET,
            url=self.repo_api_url,
            json=self._get_expected_repo(owner="TestOwner", name="TestRepo"),
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/main",
            json=self._get_expected_ref("main", "abcdef"),
        )

        source = GitHubSource(
            self.project_config,
            GitHubSourceModel(
                github="https://github.com/TestOwner/TestRepo.git", ref="main"
            ),
        )
        assert (
            repr(source) == "<GitHubSource GitHub: TestOwner/TestRepo @ main (abcdef)>"
        )

    @responses.activate
    def test_resolve__branch(self):
        responses.add(
            method=responses.GET,
            url=self.repo_api_url,
            json=self._get_expected_repo(owner="TestOwner", name="TestRepo"),
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/heads/main",
            json=self._get_expected_ref("main", "abcdef"),
        )

        source = GitHubSource(
            self.project_config,
            GitHubSourceModel(
                github="https://github.com/TestOwner/TestRepo.git", branch="main"
            ),
        )
        assert (
            repr(source) == "<GitHubSource GitHub: TestOwner/TestRepo @ main (abcdef)>"
        )

    @responses.activate
    def test_resolve__tag(self):
        responses.add(
            method=responses.GET,
            url=self.repo_api_url,
            json=self._get_expected_repo(owner="TestOwner", name="TestRepo"),
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/release/1.0",
            json=self._get_expected_tag_ref("release/1.0", "abcdef"),
        )

        source = GitHubSource(
            self.project_config,
            GitHubSourceModel(
                github="https://github.com/TestOwner/TestRepo.git", tag="release/1.0"
            ),
        )
        assert (
            repr(source)
            == "<GitHubSource GitHub: TestOwner/TestRepo @ tags/release/1.0 (abcdef)>"
        )

    @responses.activate
    def test_resolve__latest_release(self):
        responses.add(
            method=responses.GET,
            url=self.repo_api_url,
            json=self._get_expected_repo(owner="TestOwner", name="TestRepo"),
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/releases/latest",
            json=self._get_expected_release("release/1.0"),
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/release/1.0",
            json=self._get_expected_tag_ref("release/1.0", "tag_sha"),
        )

        source = GitHubSource(
            self.project_config,
            GitHubSourceModel(
                github="https://github.com/TestOwner/TestRepo.git", release="latest"
            ),
        )
        assert (
            repr(source)
            == "<GitHubSource GitHub: TestOwner/TestRepo @ tags/release/1.0 (tag_sha)>"
        )

    @responses.activate
    def test_resolve__latest_beta(self):
        responses.add(
            method=responses.GET,
            url=self.repo_api_url,
            json=self._get_expected_repo(owner="TestOwner", name="TestRepo"),
        )
        responses.add(
            "POST",
            "https://api.github.com/graphql",
            json=self._get_expected_prerelease_tag_gql("beta/1.0-Beta_1"),
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/releases/tags/beta/1.0-Beta_1",
            json=self._get_expected_release("beta/1.0-Beta_1"),
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/beta/1.0-Beta_1",
            json=self._get_expected_tag_ref("beta/1.0-Beta_1", "tag_sha"),
        )

        source = GitHubSource(
            self.project_config,
            GitHubSourceModel(
                github="https://github.com/TestOwner/TestRepo.git",
                release="latest_beta",
            ),
        )
        assert (
            repr(source)
            == "<GitHubSource GitHub: TestOwner/TestRepo @ tags/beta/1.0-Beta_1 (tag_sha)>"
        )

    @responses.activate
    def test_resolve__previous_release(self):
        responses.add(
            method=responses.GET,
            url=self.repo_api_url,
            json=self._get_expected_repo(owner="TestOwner", name="TestRepo"),
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/releases",
            json=[
                self._get_expected_release("release/2.0"),
                self._get_expected_release("beta/1.0-Beta_1", prerelease=True),
                self._get_expected_release("release/1.0"),
            ],
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/release/1.0",
            json=self._get_expected_tag_ref("release/1.0", "tag_sha"),
        )

        source = GitHubSource(
            self.project_config,
            GitHubSourceModel(
                github="https://github.com/TestOwner/TestRepo.git", release="previous"
            ),
        )
        assert (
            repr(source)
            == "<GitHubSource GitHub: TestOwner/TestRepo @ tags/release/1.0 (tag_sha)>"
        )

    @responses.activate
    def test_resolve__release_not_found(self):
        responses.add(
            method=responses.GET,
            url=self.repo_api_url,
            json=self._get_expected_repo(owner="TestOwner", name="TestRepo"),
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/releases/latest",
            status=404,
        )

        with pytest.raises(DependencyResolutionError):
            GitHubSource(
                self.project_config,
                GitHubSourceModel(
                    github="https://github.com/TestOwner/TestRepo.git", release="latest"
                ),
            )

    @responses.activate
    def test_fetch(self):
        responses.add(
            method=responses.GET,
            url=self.repo_api_url,
            json=self._get_expected_repo(owner="TestOwner", name="TestRepo"),
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/releases/latest",
            json=self._get_expected_release("release/1.0"),
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/release/1.0",
            json=self._get_expected_tag_ref("release/1.0", "tag_sha"),
        )
        f = io.BytesIO()
        zf = zipfile.ZipFile(f, "w")
        zfi = zipfile.ZipInfo("toplevel/")
        zf.writestr(zfi, "")
        zf.writestr(
            "toplevel/cumulusci.yml",
            yaml.dump(
                {
                    "project": {
                        "package": {"name_managed": "Test Product", "namespace": "ns"}
                    }
                }
            ),
        )
        zf.close()
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/zipball/tag_sha",
            body=f.getvalue(),
            content_type="application/zip",
        )

        source = GitHubSource(
            self.project_config,
            GitHubSourceModel(
                github="https://github.com/TestOwner/TestRepo.git",
                release=GitHubSourceRelease.LATEST,
            ),
        )
        project_config = source.fetch()
        assert isinstance(project_config, BaseProjectConfig)
        assert pathlib.Path(project_config.repo_root).samefile(
            os.path.join(
                self.project_config.cache_dir, "projects", "TestRepo", "tag_sha"
            )
        )

    @responses.activate
    def test_fetch__nested_sources(self):
        def mock_repo(name: str, cci_yaml: dict):
            responses.add(
                method=responses.GET,
                url=f"https://api.github.com/repos/TestOwner/{name}",
                json=self._get_expected_repo(owner="TestOwner", name=name),
            )
            responses.add(
                "GET",
                f"https://api.github.com/repos/TestOwner/{name}/releases/latest",
                json=self._get_expected_release("release/1.0"),
            )
            responses.add(
                "GET",
                f"https://api.github.com/repos/TestOwner/{name}/git/ref/tags/release/1.0",
                json=self._get_expected_tag_ref("release/1.0", "tag_sha"),
            )
            f = io.BytesIO()
            zf = zipfile.ZipFile(f, "w")
            zfi = zipfile.ZipInfo("toplevel/")
            zf.writestr(zfi, "")
            zf.writestr(
                "toplevel/cumulusci.yml",
                yaml.dump(cci_yaml),
            )
            zf.close()
            responses.add(
                "GET",
                f"https://api.github.com/repos/TestOwner/{name}/zipball/tag_sha",
                body=f.getvalue(),
                content_type="application/zip",
            )

        mock_repo(
            "Foo",
            {
                "project": {"package": {"name": "Test Foo", "namespace": "foo"}},
                "sources": {
                    "bar": {
                        "github": "https://github.com/TestOwner/Bar",
                        "release": "latest",
                    }
                },
            },
        )
        mock_repo(
            "Bar", {"project": {"package": {"name": "Test Bar", "namespace": "bar"}}}
        )

        foo_config = self.project_config.get_namespace("foo")
        assert isinstance(foo_config, BaseProjectConfig)
        bar_config = foo_config.get_namespace("bar")
        assert isinstance(bar_config, BaseProjectConfig)

        assert (
            pathlib.Path(self.project_config.cache_dir) / "projects" / "Foo" / "tag_sha"
        ).exists()
        assert (
            pathlib.Path(self.project_config.cache_dir) / "projects" / "Bar" / "tag_sha"
        ).exists()
        assert not (
            pathlib.Path(self.project_config.cache_dir)
            / "projects"
            / "Foo"
            / "tag_sha"
            / ".cci"
        ).exists()

    @responses.activate
    @mock.patch("cumulusci.vcs.vcs_source.download_extract_vcs_from_repo")
    def test_fetch__cleans_up_after_failed_extract(
        self, download_extract_vcs_from_repo
    ):
        responses.add(
            method=responses.GET,
            url=self.repo_api_url,
            json=self._get_expected_repo(owner="TestOwner", name="TestRepo"),
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/releases/latest",
            json=self._get_expected_release("release/1.0"),
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/release/1.0",
            json=self._get_expected_tag_ref("release/1.0", "tag_sha"),
        )
        # Set up a fake IOError while extracting the zipball
        download_extract_vcs_from_repo.return_value = mock.Mock(
            extractall=mock.Mock(side_effect=IOError)
        )

        source = GitHubSource(
            self.project_config,
            GitHubSourceModel(
                github="https://github.com/TestOwner/TestRepo.git",
                release=GitHubSourceRelease.LATEST,
            ),
        )
        with temporary_dir():
            with pytest.raises(IOError):
                source.fetch()
            assert not pathlib.Path(".cci", "projects", "TestRepo", "tag_sha").exists()

    @responses.activate
    def test_hash(self):
        responses.add(
            method=responses.GET,
            url=self.repo_api_url,
            json=self._get_expected_repo(owner="TestOwner", name="TestRepo"),
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/releases/latest",
            json=self._get_expected_release("release/1.0"),
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/release/1.0",
            json=self._get_expected_tag_ref("release/1.0", "tag_sha"),
        )

        source = GitHubSource(
            self.project_config,
            GitHubSourceModel(
                github="https://github.com/TestOwner/TestRepo.git", release="latest"
            ),
        )
        assert hash(source) == hash(
            ("https://github.com/TestOwner/TestRepo", "tag_sha")
        )

    @responses.activate
    def test_frozenspec(self):
        responses.add(
            method=responses.GET,
            url=self.repo_api_url,
            json=self._get_expected_repo(owner="TestOwner", name="TestRepo"),
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/releases/latest",
            json=self._get_expected_release("release/1.0"),
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/release/1.0",
            json=self._get_expected_tag_ref("release/1.0", "tag_sha"),
        )

        source = GitHubSource(
            self.project_config,
            GitHubSourceModel(
                github="https://github.com/TestOwner/TestRepo.git", release="latest"
            ),
        )
        assert source.frozenspec == {
            "vcs": "github",
            "url": "https://github.com/TestOwner/TestRepo",
            "commit": "tag_sha",
            "description": "tags/release/1.0",
        }

    @responses.activate
    def test_githubsource_init__404(self):
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo",
            status=404,
        )

        with pytest.raises(
            DependencyResolutionError, match="unable to find the repository"
        ):
            GitHubSource(
                self.project_config,
                GitHubSourceModel(
                    github="https://github.com/TestOwner/TestRepo.git", release="latest"
                ),
            )

    @responses.activate
    def test_githubsource_init__403(self):
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo",
            status=403,
            headers={"X-Github-Sso": "partial-results; organizations=0810298,20348880"},
        )

        with pytest.raises(VcsApiError):
            GitHubSource(
                self.project_config,
                GitHubSourceModel(
                    github="https://github.com/TestOwner/TestRepo.git", release="latest"
                ),
            )

    @responses.activate
    def test_resolve_github_enterprise__default(self):
        # The default is to use resolution strategy `production`, which gets the latest release.
        responses.add(
            method=responses.GET,
            url="https://git.enterprise.domain.com/api/v3/repos/TestOwner/TestRepo",
            json=self._get_expected_repo(owner="TestOwner", name="TestRepo"),
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/releases/latest",
            json=self._get_expected_release("release/1.0"),
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/release/1.0",
            json=self._get_expected_tag_ref("release/1.0", "tag_sha"),
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/tags/tag_sha",
            json=self._get_expected_tag("release/1.0", "commit_sha", "tag_sha"),
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/contents/cumulusci.yml?ref=commit_sha",
            json={
                "url": "https://api.github.com/repos/TestOwner/TestRepo/contents/cumulusci.yml?ref=commit_sha",
                "download_url": "https://api.github.com/repos/TestOwner/TestRepo/contents/cumulusci.yml?ref=commit_sha",
                "git_url": "https://api.github.com/repos/TestOwner/TestRepo/contents/cumulusci.yml?ref=commit_sha",
                "html_url": "https://api.github.com/repos/TestOwner/TestRepo/contents/cumulusci.yml?ref=commit_sha",
                "_links": {},
                "name": "cumulusci.yml",
                "path": "cumulusci.yml",
                "sha": "commit_sha",
                "size": 100,
                "type": "yaml",
                "encoding": "base64",
                "content": b64encode(
                    yaml.dump(
                        {
                            "project": {
                                "package": {
                                    "name_managed": "Test Product",
                                    "namespace": "ns",
                                }
                            }
                        }
                    ).encode("utf-8")
                ).decode("utf-8"),
            },
        )

        self.project_config.keychain.set_service(
            "github_enterprise",
            "ent",
            ServiceConfig(
                {
                    "username": "testusername",
                    "email": "test@domain.com",
                    "token": "ATOKEN",
                    "server_domain": "git.enterprise.domain.com",
                }
            ),
        )

        source = GitHubEnterpriseSource(
            self.project_config,
            GitHubSourceModel(
                github="https://git.enterprise.domain.com/TestOwner/TestRepo.git"
            ),
        )
        assert repr(source) == "<GitHubSource GitHub: TestOwner/TestRepo @ commit_sha>"


class TestLocalFolderSource:
    def test_fetch(self):
        project_config = BaseProjectConfig(
            UniversalConfig(),
            repo_info={"root": os.getcwd()},
        )
        with temporary_dir() as d:
            touch("cumulusci.yml")
            source = LocalFolderSource(project_config, LocalFolderSourceModel(path=d))
            project_config = source.fetch()
            assert project_config.repo_root == os.path.realpath(d)

    def test_hash(self):
        project_config = BaseProjectConfig(UniversalConfig())
        with temporary_dir() as d:
            source = LocalFolderSource(project_config, LocalFolderSourceModel(path=d))
            assert hash(source) == hash((source.path,))

    def test_repr(self):
        project_config = BaseProjectConfig(UniversalConfig())
        with temporary_dir() as d:
            source = LocalFolderSource(project_config, LocalFolderSourceModel(path=d))
            assert repr(source) == f"<LocalFolderSource Local folder: {d}>"

    def test_frozenspec(self):
        project_config = BaseProjectConfig(UniversalConfig())
        with temporary_dir() as d:
            source = LocalFolderSource(project_config, LocalFolderSourceModel(path=d))
            with pytest.raises(NotImplementedError):
                source.frozenspec



from unittest import mock

from cumulusci.tasks.salesforce import UninstallPackage
from cumulusci.tests.util import create_project_config

from .util import create_task


class TestUninstallPackage:
    @mock.patch(
        "cumulusci.salesforce_api.package_zip.UninstallPackageZipBuilder.__call__"
    )
    def test_get_destructive_changes(self, UninstallPackageZipBuilder):
        project_config = create_project_config()
        project_config.config["project"]["package"]["namespace"] = "ns"
        task = create_task(UninstallPackage, {}, project_config)
        task.api_class = mock.Mock()
        task._get_api()
        UninstallPackageZipBuilder.assert_called_once()



from typing import Type

from cumulusci.utils.yaml.cumulusci_yml import GitHubSourceModel, VCSSourceModel
from cumulusci.vcs.vcs_source import VCSSource


class GitHubSource(VCSSource):
    def __init__(self, project_config, spec: VCSSourceModel):
        super().__init__(project_config, spec)

    def __repr__(self):
        return f"<GitHubSource {str(self)}>"

    def __str__(self):
        s = f"GitHub: {self.repo.repo_owner}/{self.repo.repo_name}"
        if self.description:
            s += f" @ {self.description}"
        if self.commit != self.description:
            s += f" ({self.commit})"
        return s

    def __hash__(self):
        return hash((self.url, self.commit))

    @classmethod
    def source_model(self) -> Type[GitHubSourceModel]:
        return GitHubSourceModel

    def get_vcs_service(self):
        from cumulusci.vcs.github.service import get_github_service_for_url

        return get_github_service_for_url(self.project_config, self.url)

    def get_ref(self):
        return self.spec.ref

    def get_tag(self):
        return "tags/" + (self.spec.tag or "")

    def get_branch(self):
        return "heads/" + (self.spec.branch or "")

    def get_release_tag(self):
        return "tags/" + (self.spec.release or "")


class GitHubEnterpriseSource(GitHubSource):
    def get_vcs_service(self):
        """Get the GitHub Enterprise service for this source."""
        return super().get_vcs_service()



import contextlib
import io
import os
import signal
import sys
from pathlib import Path
from unittest import mock

import click
import pytest
from packaging import version
from requests.exceptions import ConnectionError
from rich.console import Console

import cumulusci
from cumulusci.cli import cci
from cumulusci.cli.tests.utils import run_click_command
from cumulusci.cli.utils import get_installed_version
from cumulusci.core.config import BaseProjectConfig
from cumulusci.core.exceptions import CumulusCIException
from cumulusci.utils import temporary_dir

MagicMock = mock.MagicMock()
CONSOLE = mock.Mock()


@pytest.fixture(autouse=True)
def reset_signal_handler_flag():
    """Reset signal handler flag to ensure clean state for tests"""
    cci._signal_handler_active = False
    yield
    cci._signal_handler_active = False


@mock.patch("cumulusci.cli.cci.tee_stdout_stderr")
@mock.patch("cumulusci.cli.cci.get_tempfile_logger")
@mock.patch("cumulusci.cli.cci.init_logger")
@mock.patch("cumulusci.cli.cci.check_latest_version")
@mock.patch("cumulusci.cli.cci.CliRuntime")
@mock.patch("cumulusci.cli.cci.cli")
def test_main(
    cli,
    CliRuntime,
    check_latest_version,
    init_logger,
    get_tempfile_logger,
    tee,
):
    get_tempfile_logger.return_value = mock.Mock(), "tempfile.log"
    cci.main()

    check_latest_version.assert_called_once()
    init_logger.assert_called_once()
    CliRuntime.assert_called_once()
    cli.assert_called_once()
    tee.assert_called_once()


@mock.patch("cumulusci.cli.cci.tee_stdout_stderr")
@mock.patch("cumulusci.cli.cci.get_tempfile_logger")
@mock.patch("cumulusci.cli.cci.init_logger")
@mock.patch("cumulusci.cli.cci.check_latest_version")
@mock.patch("cumulusci.cli.cci.CliRuntime")
@mock.patch("cumulusci.cli.cci.cli")
@mock.patch("pdb.post_mortem")
@mock.patch("sys.exit")
def test_main__debug(
    sys_exit,
    post_mortem,
    cli,
    CliRuntime,
    check_latest_version,
    init_logger,
    get_tempfile_logger,
    tee,
):
    cli.side_effect = Exception
    get_tempfile_logger.return_value = (mock.Mock(), "tempfile.log")

    cci.main(["cci", "--debug"])

    check_latest_version.assert_called_once()
    init_logger.assert_called_once_with(debug=True)
    CliRuntime.assert_called_once()
    cli.assert_called_once()
    post_mortem.assert_called_once()
    sys_exit.assert_called_once_with(1)
    get_tempfile_logger.assert_called_once()
    tee.assert_called_once()


@mock.patch("cumulusci.cli.cci.tee_stdout_stderr")
@mock.patch("cumulusci.cli.cci.get_tempfile_logger")
@mock.patch("cumulusci.cli.cci.init_logger")
@mock.patch("cumulusci.cli.cci.check_latest_version")
@mock.patch("cumulusci.cli.cci.CliRuntime")
@mock.patch("cumulusci.cli.cci.cli")
@mock.patch("pdb.post_mortem")
def test_main__cci_show_stacktraces(
    post_mortem,
    cli,
    CliRuntime,
    check_latest_version,
    init_logger,
    get_tempfile_logger,
    tee,
    capsys,
):
    runtime = mock.Mock()
    runtime.universal_config.cli__show_stacktraces = True
    CliRuntime.return_value = runtime
    cli.side_effect = Exception
    get_tempfile_logger.return_value = (mock.Mock(), "tempfile.log")

    with pytest.raises(SystemExit):
        cci.main(["cci"])

    check_latest_version.assert_called_once()
    init_logger.assert_called_once_with(debug=False)
    CliRuntime.assert_called_once()
    cli.assert_called_once()
    post_mortem.assert_not_called()
    captured = capsys.readouterr()
    assert "Traceback (most recent call last)" in captured.err


@mock.patch("cumulusci.cli.cci.tee_stdout_stderr")
@mock.patch("cumulusci.cli.cci.get_tempfile_logger")
@mock.patch("cumulusci.cli.cci.init_logger")
@mock.patch("cumulusci.cli.cci.cli")
@mock.patch("sys.exit")
def test_main__abort(
    sys_exit, cli, init_logger, get_tempfile_logger, tee_stdout_stderr
):
    get_tempfile_logger.return_value = (mock.Mock(), "tempfile.log")
    cli.side_effect = click.Abort
    cci.main(["cci"])
    cli.assert_called_once()
    sys_exit.assert_called_once_with(1)


@mock.patch("cumulusci.cli.cci.tee_stdout_stderr")
@mock.patch("cumulusci.cli.cci.get_tempfile_logger")
@mock.patch("cumulusci.cli.cci.init_logger")
@mock.patch("cumulusci.cli.cci.check_latest_version")
@mock.patch("cumulusci.cli.cci.CliRuntime")
@mock.patch("cumulusci.cli.cci.cli")
@mock.patch("pdb.post_mortem")
@mock.patch("sys.exit")
def test_main__error(
    sys_exit,
    post_mortem,
    cli,
    CliRuntime,
    check_latest_version,
    init_logger,
    get_tempfile_logger,
    tee,
):
    runtime = mock.Mock()
    runtime.universal_config.cli__show_stacktraces = False
    CliRuntime.return_value = runtime

    cli.side_effect = Exception
    get_tempfile_logger.return_value = mock.Mock(), "tempfile.log"

    cci.main(["cci", "org", "info"])

    check_latest_version.assert_called_once()
    init_logger.assert_called_once_with(debug=False)
    CliRuntime.assert_called_once()
    cli.assert_called_once()
    post_mortem.call_count == 0
    sys_exit.assert_called_once_with(1)
    get_tempfile_logger.assert_called_once()
    tee.assert_called_once()

    os.remove("tempfile.log")


@mock.patch("cumulusci.cli.cci.tee_stdout_stderr")
@mock.patch("cumulusci.cli.cci.get_tempfile_logger")
@mock.patch("cumulusci.cli.cci.CliRuntime")
def test_main__CliRuntime_error(CliRuntime, get_tempfile_logger, tee):
    CliRuntime.side_effect = CumulusCIException("something happened")
    get_tempfile_logger.return_value = mock.Mock(), "tempfile.log"

    with contextlib.redirect_stderr(io.StringIO()) as stderr:
        with mock.patch("sys.exit") as sys_exit:
            sys_exit.side_effect = SystemExit  # emulate real sys.exit
            with pytest.raises(SystemExit):
                cci.main(["cci", "org", "info"])

    assert "something happened" in stderr.getvalue()

    tempfile = Path("tempfile.log")
    tempfile.unlink()


@mock.patch("cumulusci.cli.cci.init_logger")  # side effects break other tests
@mock.patch("cumulusci.cli.cci.get_tempfile_logger")
@mock.patch("cumulusci.cli.cci.tee_stdout_stderr")
@mock.patch("cumulusci.cli.cci.CliRuntime")
@mock.patch("sys.exit", MagicMock())
def test_handle_org_name(
    CliRuntime, tee_stdout_stderr, get_tempfile_logger, init_logger
):

    # get_tempfile_logger doesn't clean up after itself which breaks other tests
    get_tempfile_logger.return_value = mock.Mock(), ""

    with contextlib.redirect_stdout(io.StringIO()) as stdout:
        cci.main(["cci", "org", "default", "xyzzy"])
    assert "xyzzy is now the default org" in stdout.getvalue()

    with contextlib.redirect_stdout(io.StringIO()) as stdout:
        cci.main(["cci", "org", "default", "--org", "xyzzy2"])
    assert "xyzzy2 is now the default org" in stdout.getvalue()

    with contextlib.redirect_stderr(io.StringIO()) as stderr:
        cci.main(["cci", "org", "default", "xyzzy1", "--org", "xyzzy2"])
    assert "not both" in stderr.getvalue()

    CliRuntime().keychain.get_default_org.return_value = ("xyzzy3", None)

    # cci org remove should really need an attached org
    with contextlib.redirect_stderr(io.StringIO()) as stderr:
        cci.main(["cci", "org", "remove"])
    assert "Please specify ORGNAME or --org ORGNAME" in stderr.getvalue()


@mock.patch("cumulusci.cli.cci.init_logger")  # side effects break other tests
@mock.patch("cumulusci.cli.cci.get_tempfile_logger")
@mock.patch("cumulusci.cli.cci.tee_stdout_stderr")
@mock.patch("sys.exit")
@mock.patch("cumulusci.cli.cci.CliRuntime")
def test_cci_org_default__no_orgname(
    CliRuntime, exit, tee_stdout_stderr, get_tempfile_logger, init_logger
):
    # get_tempfile_logger doesn't clean up after itself which breaks other tests
    get_tempfile_logger.return_value = mock.Mock(), ""

    CliRuntime().keychain.get_default_org.return_value = ("xyzzy4", None)
    with contextlib.redirect_stdout(io.StringIO()) as stdout:
        cci.main(["cci", "org", "default"])
    assert "xyzzy4 is the default org" in stdout.getvalue()

    CliRuntime().keychain.get_default_org.return_value = (None, None)
    with contextlib.redirect_stdout(io.StringIO()) as stdout:
        cci.main(["cci", "org", "default"])
    assert "There is no default org" in stdout.getvalue()


DEPLOY_CLASS_PATH = f"cumulusci.tasks.salesforce.Deploy{'.Deploy' if sys.version_info >= (3, 11) else ''}"


@mock.patch(
    "cumulusci.cli.cci.init_logger", mock.Mock()
)  # side effects break other tests
@mock.patch("cumulusci.cli.cci.tee_stdout_stderr", mock.MagicMock())
@mock.patch(f"{DEPLOY_CLASS_PATH}.__call__", mock.Mock())
@mock.patch("sys.exit", mock.Mock())
@mock.patch("cumulusci.cli.cci.get_tempfile_logger")
@mock.patch("cumulusci.cli.cci.CliRuntime")
@mock.patch(f"{DEPLOY_CLASS_PATH}.__init__")
def test_cci_run_task_options__with_dash(
    Deploy,
    CliRuntime,
    get_tempfile_logger,
):
    # get_tempfile_logger doesn't clean up after itself which breaks other tests
    Deploy.return_value = None
    get_tempfile_logger.return_value = mock.Mock(), ""
    CliRuntime.return_value = runtime = mock.Mock()
    runtime.get_org.return_value = ("test", mock.Mock())
    runtime.project_config = BaseProjectConfig(
        runtime.universal_config,
        {
            "project": {"name": "Test"},
            "tasks": {"deploy": {"class_path": "cumulusci.tasks.salesforce.Deploy"}},
        },
    )

    cci.main(
        ["cci", "task", "run", "deploy", "--path", "x", "--clean-meta-xml", "False"]
    )
    task_config = Deploy.mock_calls[0][1][1]
    assert "clean_meta_xml" in task_config.options


@mock.patch("cumulusci.cli.cci.init_logger", mock.Mock())
@mock.patch("cumulusci.cli.cci.tee_stdout_stderr", mock.MagicMock())
@mock.patch(f"{DEPLOY_CLASS_PATH}.__call__", mock.Mock())
@mock.patch("sys.exit", mock.Mock())
@mock.patch("cumulusci.cli.cci.get_tempfile_logger")
@mock.patch("cumulusci.cli.cci.CliRuntime")
@mock.patch(f"{DEPLOY_CLASS_PATH}.__init__")
def test_cci_run_task_options__old_style_with_dash(
    Deploy,
    CliRuntime,
    get_tempfile_logger,
):
    # get_tempfile_logger doesn't clean up after itself which breaks other tests
    Deploy.return_value = None
    get_tempfile_logger.return_value = mock.Mock(), ""
    CliRuntime.return_value = runtime = mock.Mock()
    runtime.get_org.return_value = ("test", mock.Mock())
    runtime.project_config = BaseProjectConfig(
        runtime.universal_config,
        {
            "project": {"name": "Test"},
            "tasks": {"deploy": {"class_path": "cumulusci.tasks.salesforce.Deploy"}},
        },
    )

    cci.main(
        [
            "cci",
            "task",
            "run",
            "deploy",
            "--path",
            "x",
            "-o",
            "clean-meta-xml",
            "False",
        ]
    )
    task_config = Deploy.mock_calls[0][1][1]
    assert "clean_meta_xml" in task_config.options


@mock.patch("cumulusci.cli.cci.open")
@mock.patch("cumulusci.cli.cci.traceback")
def test_handle_exception(traceback, cci_open):
    console = mock.Mock()
    Console.return_value = console
    error_message = "foo"
    cci_open.__enter__.return_value = mock.Mock()

    with contextlib.redirect_stderr(io.StringIO()) as stderr:
        cci.handle_exception(error_message, False, "logfile.path")

    stderr = stderr.getvalue()
    assert f"Error: {error_message}" in stderr
    assert cci.SUGGEST_ERROR_COMMAND in stderr
    traceback.print_exc.assert_called_once()


@mock.patch("cumulusci.cli.cci.open")
def test_handle_exception__error_cmd(cci_open):
    """Ensure we don't write to logfiles when running `cci error ...` commands."""
    error_message = "foo"
    logfile_path = None

    with contextlib.redirect_stderr(io.StringIO()) as stderr:
        cci.handle_exception(error_message, False, logfile_path)

    stderr = stderr.getvalue()
    assert f"Error: {error_message}" in stderr
    assert cci.SUGGEST_ERROR_COMMAND in stderr
    cci_open.assert_not_called()


@mock.patch("cumulusci.cli.cci.open")
@mock.patch("cumulusci.cli.cci.traceback")
def test_handle_click_exception(traceback, cci_open):
    cci_open.__enter__.return_value = mock.Mock()

    with contextlib.redirect_stderr(io.StringIO()) as stderr:
        cci.handle_exception(click.ClickException("[oops]"), False, "file.path")

    stderr = stderr.getvalue()
    assert "Error: [oops]" in stderr
    traceback.assert_not_called()


@mock.patch("cumulusci.cli.cci.open")
def test_handle_connection_exception(cci_open):
    cci_open.__enter__.return_value = mock.Mock()

    with contextlib.redirect_stderr(io.StringIO()) as stderr:
        cci.handle_exception(ConnectionError(), False, "file.log")

    stderr = stderr.getvalue()
    assert "We encountered an error with your internet connection." in stderr


def test_cli():
    run_click_command(cci.cli)


@mock.patch(
    "cumulusci.cli.cci.get_latest_final_version",
    mock.Mock(return_value=version.parse("100")),
)
def test_version(capsys):
    run_click_command(cci.version)
    console_output = capsys.readouterr().out
    assert f"CumulusCI Plus version: {cumulusci.__version__}" in console_output
    assert "There is a newer version of CumulusCI Plus available" in console_output


@mock.patch(
    "cumulusci.cli.cci.get_latest_final_version",
    mock.Mock(return_value=version.parse("1")),
)
def test_version__latest(capsys):
    run_click_command(cci.version)
    console_output = capsys.readouterr().out
    assert "You have the latest version of CumulusCI" in console_output


@mock.patch("cumulusci.cli.cci.warn_if_no_long_paths")
@mock.patch("cumulusci.cli.cci.get_latest_final_version", get_installed_version)
def test_version__win_path_warning(warn_if):
    run_click_command(cci.version)
    warn_if.assert_called_once()


@mock.patch("code.interact")
def test_shell(interact):
    run_click_command(cci.shell)
    interact.assert_called_once()
    assert "config" in interact.call_args[1]["local"]
    assert "runtime" in interact.call_args[1]["local"]


@mock.patch("runpy.run_path")
def test_shell_script(runpy):
    run_click_command(cci.shell, script="foo.py")
    runpy.assert_called_once()
    assert "config" in runpy.call_args[1]["init_globals"]
    assert "runtime" in runpy.call_args[1]["init_globals"]
    assert runpy.call_args[0][0] == "foo.py", runpy.call_args[0]


@mock.patch("builtins.print")
def test_shell_code(print):
    run_click_command(cci.shell, python="print(config, runtime)")
    print.assert_called_once()


@mock.patch("cumulusci.cli.cci.print")
def test_shell_mutually_exclusive_args(print):
    with pytest.raises(Exception) as e:
        run_click_command(cci.shell, script="foo.py", python="print(config, runtime)")
    assert "Cannot specify both" in e.value.message


@mock.patch("code.interact")
def test_shell__no_project(interact):
    with temporary_dir():
        run_click_command(cci.shell)
        interact.assert_called_once()


def test_cover_command_groups():
    run_click_command(cci.project)
    run_click_command(cci.org)
    run_click_command(cci.task)
    run_click_command(cci.flow)
    run_click_command(cci.service)
    # no assertion; this test is for coverage of empty methods


@mock.patch(
    "cumulusci.cli.runtime.CliRuntime.get_org",
    lambda *args, **kwargs: (MagicMock(), MagicMock()),
)
@mock.patch("cumulusci.core.runtime.BaseCumulusCI._load_keychain", MagicMock())
@mock.patch("pdb.post_mortem", MagicMock())
@mock.patch("cumulusci.cli.cci.tee_stdout_stderr", MagicMock())
@mock.patch("cumulusci.cli.cci.init_logger", MagicMock())
@mock.patch("cumulusci.cli.cci.get_tempfile_logger")
def test_run_task_debug(get_tempfile_logger):
    get_tempfile_logger.return_value = (mock.Mock(), "tempfile.log")

    gipnew = "cumulusci.tasks.preflight.packages.GetInstalledPackages._run_task"
    with mock.patch(gipnew, mock_validate_debug(False)):
        cci.main(["cci", "task", "run", "get_installed_packages"])
    with mock.patch(gipnew, mock_validate_debug(True)):
        cci.main(["cci", "task", "run", "get_installed_packages", "--debug"])


@mock.patch(
    "cumulusci.cli.runtime.CliRuntime.get_org",
    lambda *args, **kwargs: (MagicMock(), MagicMock()),
)
@mock.patch("cumulusci.core.runtime.BaseCumulusCI._load_keychain", MagicMock())
@mock.patch("pdb.post_mortem", MagicMock())
@mock.patch("cumulusci.cli.cci.tee_stdout_stderr", MagicMock())
@mock.patch("cumulusci.cli.cci.init_logger", MagicMock())
@mock.patch("cumulusci.tasks.robotframework.RobotLibDoc", MagicMock())
@mock.patch("cumulusci.cli.cci.get_tempfile_logger")
def test_run_flow_debug(get_tempfile_logger):
    get_tempfile_logger.return_value = (mock.Mock(), "tempfile.log")
    rtd = "cumulusci.tasks.robotframework.RobotTestDoc._run_task"

    with mock.patch(rtd, mock_validate_debug(False)):
        cci.main(["cci", "flow", "run", "robot_docs"])
    with mock.patch(rtd, mock_validate_debug(True)):
        cci.main(["cci", "flow", "run", "robot_docs", "--debug"])


def mock_validate_debug(value):
    def _run_task(self, *args, **kwargs):
        assert bool(self.debug_mode) == bool(value)

    return _run_task


@mock.patch("cumulusci.cli.cci.tee_stdout_stderr")
@mock.patch("cumulusci.cli.cci.get_tempfile_logger")
@mock.patch("cumulusci.cli.cci.init_logger")
@mock.patch("cumulusci.cli.cci.check_latest_version")
@mock.patch("cumulusci.cli.cci.CliRuntime")
@mock.patch("cumulusci.cli.cci.show_version_info")
def test_dash_dash_version(
    show_version_info,
    CliRuntime,
    check_latest_version,
    init_logger,
    get_tempfile_logger,
    tee,
):
    get_tempfile_logger.return_value = mock.Mock(), "tempfile.log"
    cci.main(["cci", "--help"])
    assert len(show_version_info.mock_calls) == 0

    cci.main(["cci", "version"])
    assert len(show_version_info.mock_calls) == 1

    cci.main(["cci", "--version"])
    assert len(show_version_info.mock_calls) == 2


@mock.patch("sys.exit")
@mock.patch("cumulusci.cli.cci.Console")
@mock.patch("os.killpg", create=True)
@mock.patch("os.getpgrp", create=True)
@mock.patch("signal.signal")
def test_signal_handler_terminates_process_group(
    mock_signal, mock_getpgrp, mock_killpg, mock_console, mock_exit
):
    """Test that the signal handler terminates the process group"""
    console_instance = mock_console.return_value
    mock_getpgrp.return_value = 1234  # Mock process group ID

    # Mock the global exit stack
    mock_exit_stack = mock.Mock()
    with mock.patch.object(cci, "_exit_stack", mock_exit_stack):
        # Mock hasattr to return True for Unix functions
        with mock.patch("builtins.hasattr", return_value=True):
            # Call the signal handler with SIGTERM
            cci._signal_handler(signal.SIGTERM, None)

            # Verify console output
            console_instance.print.assert_any_call(
                "\n[yellow]Received SIGTERM - CumulusCI is being terminated[/yellow]"
            )
            console_instance.print.assert_any_call(
                "[yellow]Exiting with failure code due to external cancellation.[/yellow]"
            )
            console_instance.print.assert_any_call(
                "[yellow]Terminating child processes...[/yellow]"
            )

            # Verify cleanup was called
            mock_exit_stack.close.assert_called_once()

            # Verify signal was temporarily ignored and then restored
            mock_signal.assert_called()

            # Verify process group was terminated
            mock_getpgrp.assert_called_once()
            mock_killpg.assert_called_once_with(1234, signal.SIGTERM)

            # Verify exit with correct code
            mock_exit.assert_called_once_with(143)


@mock.patch("sys.exit")
@mock.patch("cumulusci.cli.cci.Console")
@mock.patch("os.killpg", create=True)
@mock.patch("os.getpgrp", create=True)
@mock.patch("signal.signal")
def test_signal_handler_sigint(
    mock_signal, mock_getpgrp, mock_killpg, mock_console, mock_exit
):
    """Test that the signal handler properly handles SIGINT"""
    console_instance = mock_console.return_value
    mock_getpgrp.return_value = 1234

    # Mock the global exit stack
    mock_exit_stack = mock.Mock()
    with mock.patch.object(cci, "_exit_stack", mock_exit_stack):
        # Mock hasattr to return True for Unix functions
        with mock.patch("builtins.hasattr", return_value=True):
            # Call the signal handler with SIGINT
            cci._signal_handler(signal.SIGINT, None)

            # Verify console output
            console_instance.print.assert_any_call(
                "\n[yellow]Received SIGINT - CumulusCI is being terminated[/yellow]"
            )

            # Verify process group was terminated with SIGINT
            mock_killpg.assert_called_once_with(1234, signal.SIGINT)

            # Verify exit with correct code for SIGINT
            mock_exit.assert_called_once_with(130)


@mock.patch("sys.exit")
@mock.patch("cumulusci.cli.cci.Console")
def test_signal_handler_prevents_recursion(mock_console, mock_exit):
    """Test that the signal handler prevents recursive calls"""
    console_instance = mock_console.return_value

    # Set the flag to simulate handler already active
    with mock.patch.object(cci, "_signal_handler_active", True):
        # Call the signal handler
        cci._signal_handler(signal.SIGTERM, None)

        # Verify no console output (handler should return immediately)
        console_instance.print.assert_not_called()

        # Verify no exit call
        mock_exit.assert_not_called()


@mock.patch("sys.exit")
@mock.patch("cumulusci.cli.cci.Console")
@mock.patch("os.killpg", create=True)
@mock.patch("os.getpgrp", create=True)
@mock.patch("signal.signal")
def test_signal_handler_handles_killpg_error(
    mock_signal, mock_getpgrp, mock_killpg, mock_console, mock_exit
):
    """Test that the signal handler handles errors from killpg gracefully"""
    console_instance = mock_console.return_value
    mock_getpgrp.return_value = 1234
    mock_killpg.side_effect = OSError("Process group not found")

    # Mock hasattr to return True for Unix functions
    with mock.patch("builtins.hasattr", return_value=True):
        # Call the signal handler with SIGTERM
        cci._signal_handler(signal.SIGTERM, None)

    # Verify error message was printed
    console_instance.print.assert_any_call(
        "[red]Warning: Error terminating child processes: Process group not found[/red]"
    )

    # Verify it still exits with correct code
    mock_exit.assert_called_once_with(143)


@mock.patch("sys.exit")
@mock.patch("cumulusci.cli.cci.Console")
def test_signal_handler_without_process_group_support(mock_console, mock_exit):
    """Test that the signal handler works on Windows where process groups aren't supported"""
    console_instance = mock_console.return_value

    # Mock the global exit stack
    mock_exit_stack = mock.Mock()
    with mock.patch.object(cci, "_exit_stack", mock_exit_stack):
        # Mock hasattr to return False for Unix functions (like on Windows)
        with mock.patch("builtins.hasattr", return_value=False):
            # Call the signal handler with SIGTERM
            cci._signal_handler(signal.SIGTERM, None)

            # Verify console output
            console_instance.print.assert_any_call(
                "\n[yellow]Received SIGTERM - CumulusCI is being terminated[/yellow]"
            )
            console_instance.print.assert_any_call(
                "[yellow]Exiting with failure code due to external cancellation.[/yellow]"
            )
            console_instance.print.assert_any_call(
                "[yellow]Terminating child processes...[/yellow]"
            )
            console_instance.print.assert_any_call(
                "[yellow]Process group termination not supported on this platform[/yellow]"
            )

            # Verify cleanup was called
            mock_exit_stack.close.assert_called_once()

            # Verify exit with correct code
            mock_exit.assert_called_once_with(143)


@mock.patch("os.setpgrp", create=True)
def test_main_creates_process_group(mock_setpgrp):
    """Test that main() creates a new process group"""
    # Mock dependencies to avoid actual CLI execution
    with mock.patch.multiple(
        "cumulusci.cli.cci",
        signal=mock.Mock(),
        init_requests_trust=mock.Mock(),
        check_latest_version=mock.Mock(),
        check_latest_plugins=mock.Mock(),
        get_tempfile_logger=mock.Mock(return_value=(mock.Mock(), "tempfile.log")),
        tee_stdout_stderr=mock.Mock(
            return_value=mock.Mock(__enter__=mock.Mock(), __exit__=mock.Mock())
        ),
        set_debug_mode=mock.Mock(
            return_value=mock.Mock(__enter__=mock.Mock(), __exit__=mock.Mock())
        ),
        CliRuntime=mock.Mock(),
        init_logger=mock.Mock(),
        cli=mock.Mock(),
    ):
        try:
            cci.main(["cci", "version"])
        except SystemExit:
            pass  # Expected for version command

        # Verify process group was created
        mock_setpgrp.assert_called_once()


@mock.patch("os.setpgrp", create=True)
def test_main_handles_setpgrp_error(mock_setpgrp):
    """Test that main() handles setpgrp errors gracefully"""
    mock_setpgrp.side_effect = OSError("Operation not permitted")

    # Mock dependencies to avoid actual CLI execution
    with mock.patch.multiple(
        "cumulusci.cli.cci",
        signal=mock.Mock(),
        init_requests_trust=mock.Mock(),
        check_latest_version=mock.Mock(),
        check_latest_plugins=mock.Mock(),
        get_tempfile_logger=mock.Mock(return_value=(mock.Mock(), "tempfile.log")),
        tee_stdout_stderr=mock.Mock(
            return_value=mock.Mock(__enter__=mock.Mock(), __exit__=mock.Mock())
        ),
        set_debug_mode=mock.Mock(
            return_value=mock.Mock(__enter__=mock.Mock(), __exit__=mock.Mock())
        ),
        CliRuntime=mock.Mock(),
        init_logger=mock.Mock(),
        cli=mock.Mock(),
    ):
        try:
            # Should not raise an exception even if setpgrp fails
            cci.main(["cci", "version"])
        except SystemExit:
            pass  # Expected for version command
        except OSError:
            pytest.fail("main() should handle setpgrp errors gracefully")

        # Verify setpgrp was attempted
        mock_setpgrp.assert_called_once()


@mock.patch("signal.signal")
def test_main_registers_signal_handlers(mock_signal):
    """Test that main() registers signal handlers"""
    # Create a proper context manager mock
    context_manager_mock = mock.Mock()
    context_manager_mock.__enter__ = mock.Mock(return_value=context_manager_mock)
    context_manager_mock.__exit__ = mock.Mock(return_value=None)

    # Create another context manager mock for set_debug_mode
    debug_context_mock = mock.Mock()
    debug_context_mock.__enter__ = mock.Mock(return_value=debug_context_mock)
    debug_context_mock.__exit__ = mock.Mock(return_value=None)

    # Mock dependencies to avoid actual CLI execution
    with mock.patch.multiple(
        "cumulusci.cli.cci",
        init_requests_trust=mock.Mock(),
        check_latest_version=mock.Mock(),
        check_latest_plugins=mock.Mock(),
        get_tempfile_logger=mock.Mock(return_value=(mock.Mock(), "tempfile.log")),
        tee_stdout_stderr=mock.Mock(return_value=context_manager_mock),
        set_debug_mode=mock.Mock(return_value=debug_context_mock),
        CliRuntime=mock.Mock(),
        init_logger=mock.Mock(),
        cli=mock.Mock(),
    ):
        try:
            cci.main(["cci", "version"])
        except SystemExit:
            pass  # Expected for version command

        # Verify signal handlers were registered
        expected_calls = [
            mock.call(signal.SIGTERM, cci._signal_handler),
            mock.call(signal.SIGINT, cci._signal_handler),
        ]
        mock_signal.assert_has_calls(expected_calls, any_order=True)



""" Tasks for running a command in a subprocess

Command - run a command with optional environment variables
SalesforceCommand - run a command with credentials passed
SalesforceBrowserTest - a task designed to wrap browser testing that could run locally or remotely
"""

import json
import os
import sys

import sarge

from cumulusci.core.exceptions import CommandException
from cumulusci.core.tasks import BaseTask
from cumulusci.core.utils import process_bool_arg


class Command(BaseTask):
    """Execute a shell command in a subprocess"""

    task_docs = """
        **Example Command-line Usage:**
        ``cci task run command -o command "echo 'Hello command task!'"``

        **Example Task to Run Command:**

        .. code-block:: yaml

            hello_world:
                description: Says hello world
                class_path: cumulusci.tasks.command.Command
                options:
                    command: echo 'Hello World!'
"""

    task_options = {
        "command": {"description": "The command to execute", "required": True},
        "dir": {
            "description": "If provided, the directory where the command "
            "should be run from."
        },
        "env": {
            "description": "Environment variables to set for command. Must "
            "be flat dict, either as python dict from YAML or "
            "as JSON string."
        },
        "pass_env": {
            "description": "If False, the current environment variables "
            "will not be passed to the child process. "
            "Defaults to True",
            "required": True,
        },
        "interactive": {
            "description": "If True, the command will use stderr, stdout, "
            "and stdin of the main process."
            "Defaults to False."
        },
    }

    def _init_options(self, kwargs):
        super(Command, self)._init_options(kwargs)
        if "pass_env" not in self.options:
            self.options["pass_env"] = True
        if "dir" not in self.options or not self.options["dir"]:
            self.options["dir"] = "."
        if "interactive" not in self.options:
            self.options["interactive"] = False
        if "env" not in self.options:
            self.options["env"] = {}
        else:
            try:
                self.options["env"] = json.loads(self.options["env"])
            except TypeError:
                # assume env is already dict
                pass

    def _run_task(self):
        env = self._get_env()
        self._run_command(env)

    def _get_env(self):
        if process_bool_arg(self.options["pass_env"]):
            env = os.environ.copy()
        else:
            env = {}

        env.update(self.options["env"])
        return env

    def _process_output(self, line):
        self.logger.info(line.decode("utf-8").rstrip())

    def _handle_returncode(self, returncode, stderr):
        if returncode:
            message = "Return code: {}".format(returncode)
            if stderr:
                message += "\nstderr: {}".format(stderr.read().decode("utf-8"))
            self.logger.error(message)
            raise CommandException(message)

    def _get_command(self):
        return self.options["command"]

    def _run_command(
        self, env, command=None, output_handler=None, return_code_handler=None
    ):
        if not command:
            command = self._get_command()

        interactive_mode = process_bool_arg(self.options["interactive"])

        self.logger.info(f"Running command: {command}")

        p = sarge.Command(
            command,
            stdout=sys.stdout if interactive_mode else sarge.Capture(buffer_size=-1),
            stderr=sys.stderr if interactive_mode else sarge.Capture(buffer_size=-1),
            shell=True,
            env=env,
            cwd=self.options.get("dir"),
        )
        if interactive_mode:
            p.run(input=sys.stdin)
        else:
            p.run(async_=True)
            # Handle output lines
            if not output_handler:
                output_handler = self._process_output
            while True:
                line = p.stdout.readline(timeout=1.0)
                if line:
                    output_handler(line)
                elif p.poll() is not None:
                    break
            p.wait()
        self.return_values = {"returncode": p.returncode}

        # Handle return code
        if not return_code_handler:
            return_code_handler = self._handle_returncode
        return_code_handler(p.returncode, None if interactive_mode else p.stderr)


class SalesforceCommand(Command):
    """Execute a Command with SF credentials provided on the environment.

    Provides:
     * SF_ORG_INSTANCE_URL
     * SF_ACCESS_TOKEN
    """

    salesforce_task = True

    def _update_credentials(self):
        self.org_config.refresh_oauth_token(self.project_config.keychain)

    def _get_env(self):
        env = super(SalesforceCommand, self)._get_env()
        env["SF_ACCESS_TOKEN"] = self.org_config.access_token
        env["SF_ORG_INSTANCE_URL"] = self.org_config.instance_url
        return env


task_options = Command.task_options.copy()
task_options["extra"] = {
    "description": "If provided, will be appended to the end of the "
    "command.  Use to pass extra args to the command.",
    "required": False,
}
task_options["use_saucelabs"] = {
    "description": "If True, use SauceLabs to run the tests. The "
    "SauceLabs credentials will be fetched from the "
    "saucelabs service in the keychain and passed as "
    "environment variables to the command.  Defaults to "
    "False to run tests in the local browser.",
    "required": True,
}



import os
from pathlib import Path
from typing import Optional

from cumulusci.core.config import BaseTaskFlowConfig
from cumulusci.core.config.project_config import (
    BaseProjectConfig,
    ProjectConfigPropertiesMixin,
)
from cumulusci.utils.yaml.cumulusci_yml import cci_safe_load

__location__ = os.path.dirname(os.path.realpath(__file__))


class UniversalConfig(BaseTaskFlowConfig, ProjectConfigPropertiesMixin):
    """Base class for the global config which contains all configuration not specific to projects"""

    project_local_dir: str
    cli: dict

    config: Optional[dict] = None
    config_filename = "cumulusci.yml"
    project_config_class = BaseProjectConfig

    def __init__(self, config=None):
        self._init_logger()
        self._load_config()

    @property
    def cumulusci_config_dir(self):
        """Get the root directory for storing persistent data, as an instance property"""
        return self.default_cumulusci_dir()

    @staticmethod
    def default_cumulusci_dir():
        """Get the root directory for storing persistent data (~/.cumulusci)

        Creates it if it doesn't exist yet.
        """
        config_dir = Path.home() / ".cumulusci"

        if not config_dir.exists():
            config_dir.mkdir(parents=True)

        return config_dir

    @property
    def config_global_path(self):
        """The global config path. Usually ~/.cumulusci/cumulusci.yml"""
        directory = self.cumulusci_config_dir
        if not os.path.exists(directory):
            os.makedirs(directory)

        config_path = os.path.join(directory, self.config_filename)
        if not os.path.isfile(config_path):
            return None

        return config_path

    @property
    def config_universal_path(self):
        return os.path.abspath(
            os.path.join(__location__, "..", "..", self.config_filename)
        )

    def _load_config(self):
        """Loads the local configuration"""
        # avoid loading multiple times
        if UniversalConfig.config is not None:
            return

        # load the universal config
        UniversalConfig.config_universal = cci_safe_load(self.config_universal_path)

        # Load the local config
        if self.config_global_path:
            config = cci_safe_load(self.config_global_path)
        else:
            config = {}
        UniversalConfig.config_global = config

        UniversalConfig.config = self.merge_base_config(
            {
                "universal_config": UniversalConfig.config_universal,
                "global_config": UniversalConfig.config_global,
            }
        )



import io
from collections import defaultdict
from unittest.mock import Mock, call, mock_open, patch

import pytest

from cumulusci.core.config.project_config import BaseProjectConfig
from cumulusci.core.dependencies import parse_dependencies
from cumulusci.core.dependencies.github import GitHubDynamicDependency
from cumulusci.core.exceptions import (
    DependencyParseError,
    DependencyResolutionError,
    TaskOptionsError,
)
from cumulusci.tasks.datadictionary import (
    PRERELEASE_SIGIL,
    FieldDetail,
    GenerateDataDictionary,
    Package,
    PackageVersion,
    SObjectDetail,
)
from cumulusci.tasks.salesforce.tests.util import create_task
from cumulusci.tests.util import create_project_config
from cumulusci.utils import temporary_dir
from cumulusci.utils.version_strings import LooseVersion
from cumulusci.utils.xml import metadata_tree
from cumulusci.utils.yaml.cumulusci_yml import cci_safe_load


class TestGenerateDataDictionary:
    def test_version_from_tag_name(self):
        task = create_task(GenerateDataDictionary, {})

        assert task._version_from_tag_name("release/1.1", "release/") == LooseVersion(
            "1.1"
        )

    def test_write_object_results(self):
        task = create_task(GenerateDataDictionary, {})

        p = Package(
            repo=None, package_name="Test", namespace="test__", prefix_release="rel/"
        )
        v = PackageVersion(package=p, version=LooseVersion("1.1"))
        v2 = PackageVersion(package=p, version=LooseVersion("1.2"))
        task.package_versions = defaultdict(list)
        task.package_versions[p] = [v2.version, v.version]
        task.sobjects = defaultdict(list)
        task.sobjects["test__Test__c"] = [
            SObjectDetail(
                version=v,
                api_name="test__Test__c",
                label="Test",
                description="Description",
            )
        ]

        f = io.StringIO()
        task._write_object_results(f)

        f.seek(0)
        result = f.read()

        assert (
            result
            == '"Object Label","Object API Name","Object Description","Version Introduced","Version Deleted"\r\n"Test","test__Test__c","Description","Test 1.1","Test 1.2"\r\n'
        )

    def test_write_field_results(self):
        task = create_task(GenerateDataDictionary, {})

        p = Package(
            repo=None, package_name="Test", namespace="test__", prefix_release="rel/"
        )
        v = PackageVersion(package=p, version=LooseVersion("1.1"))
        v2 = PackageVersion(package=p, version=LooseVersion("1.2"))

        task._init_schema()
        task.package_versions[p] = [v2.version, v.version]
        task.sobjects["test__Test__c"] = [
            SObjectDetail(
                version=v,
                api_name="test__Test__c",
                label="Test Object",
                description="Desc",
            ),
            SObjectDetail(
                version=v2,
                api_name="test__Test__c",
                label="Test Object",
                description="Desc",
            ),
        ]

        task.fields.update(
            {
                "Account.test__Desc__c": [
                    FieldDetail(
                        version=v2,
                        sobject="Account",
                        api_name="test__Desc__c",
                        label="Desc",
                        type="Text",
                        help_text="",
                        description="",
                        valid_values="",
                    )
                ],
                "test__Test__c.test__Type__c": [
                    FieldDetail(
                        version=v,
                        sobject="test__Test__c",
                        api_name="test__Type__c",
                        label="Type",
                        type="Picklist",
                        help_text="Help",
                        description="Description",
                        valid_values="Foo; Bar",
                    ),
                    FieldDetail(
                        version=v2,
                        sobject="test__Test__c",
                        api_name="test__Type__c",
                        label="Type",
                        type="Picklist",
                        help_text="New Help",
                        description="Description",
                        valid_values="Foo; Bar; New Value",
                    ),
                ],
                "test__Test__c.test__Account__c": [
                    FieldDetail(
                        version=v,
                        sobject="test__Test__c",
                        api_name="test__Account__c",
                        label="Account",
                        type="Lookup to Account",
                        help_text="Help",
                        description="Description",
                        valid_values="",
                    )
                ],
            }
        )

        f = io.StringIO()
        task._write_field_results(f)
        f.seek(0)
        result = f.read()

        assert result == (
            '"Object Label","Object API Name","Field Label","Field API Name",'
            '"Type","Picklist Values","Help Text","Field Description","Version Introduced",'
            '"Version Picklist Values Last Changed","Version Help Text Last Changed",'
            '"Version Deleted"\r\n"Account","Account","Desc","test__Desc__c","Text",'
            '"","","","Test 1.2","","",""\r\n"Test Object","test__Test__c","Type",'
            '"test__Type__c","Picklist","Foo; Bar; New Value","New Help","Description",'
            '"Test 1.1","Test 1.2","Test 1.2",""\r\n"Test Object","test__Test__c",'
            '"Account","test__Account__c","Lookup to Account","","Help","Description",'
            '"Test 1.1","","","Test 1.2"\r\n'
        )

    def test_write_field_results__omit_sobjects(self):
        task = create_task(GenerateDataDictionary, {})

        p = Package(
            repo=None, package_name="Test", namespace="test__", prefix_release="rel/"
        )
        v = PackageVersion(package=p, version=LooseVersion("1.1"))
        v2 = PackageVersion(package=p, version=LooseVersion("1.2"))
        task._init_schema()
        task.package_versions[p] = [v2.version, v.version]
        task.omit_sobjects = set(["test__Test2__c"])
        task.sobjects["test__Test__c"] = [
            SObjectDetail(
                version=v,
                api_name="test__Test__c",
                label="Test Object",
                description="Desc",
            ),
            SObjectDetail(
                version=v2,
                api_name="test__Test__c",
                label="Test Object",
                description="Desc",
            ),
        ]
        task.fields["test__Test2__c.test__Blah__c"] = [
            FieldDetail(
                version=v,
                sobject="test__Test2__c",
                api_name="test__Blah__c",
                label="Test Field",
                type="Text",
                help_text="Help",
                description="Description",
                valid_values="",
            )
        ]
        task.fields["test__Test__c.test__Type__c"] = [
            FieldDetail(
                version=v,
                sobject="test__Test__c",
                api_name="test__Type__c",
                label="Type",
                type="Picklist",
                help_text="Help",
                description="Description",
                valid_values="Foo; Bar",
            ),
            FieldDetail(
                version=v2,
                sobject="test__Test__c",
                api_name="test__Type__c",
                label="Type",
                type="Picklist",
                help_text="New Help",
                description="Description",
                valid_values="Foo; Bar; New Value",
            ),
        ]
        task.fields["test__Test__c.test__Account__c"] = [
            FieldDetail(
                version=v,
                sobject="test__Test__c",
                api_name="test__Account__c",
                label="Account",
                type="Lookup to Account",
                help_text="Help",
                description="Description",
                valid_values="",
            )
        ]

        f = io.StringIO()
        task._write_field_results(f)
        f.seek(0)
        result = f.read()

        assert result == (
            '"Object Label","Object API Name","Field Label","Field API Name","Type",'
            '"Picklist Values","Help Text","Field Description","Version Introduced",'
            '"Version Picklist Values Last Changed","Version Help Text Last Changed",'
            '"Version Deleted"\r\n"Test Object","test__Test__c","Type","test__Type__c",'
            '"Picklist","Foo; Bar; New Value","New Help","Description","Test 1.1","Test 1.2",'
            '"Test 1.2",""\r\n"Test Object","test__Test__c","Account","test__Account__c",'
            '"Lookup to Account","","Help","Description","Test 1.1","","","Test 1.2"\r\n'
        )

    def test_should_process_object(self):
        object_source_negative = b"""<?xml version="1.0" encoding="UTF-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
    <description>Description</description>
    <label>Test</label>
    <visibility>Protected</visibility>
</CustomObject>"""

        object_source_positive = b"""<?xml version="1.0" encoding="UTF-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
    <description>Description</description>
    <label>Test</label>
</CustomObject>"""

        task = create_task(GenerateDataDictionary, {})

        assert task._should_process_object(
            "test__", "test__Obj__c", metadata_tree.fromstring(object_source_positive)
        )
        assert task._should_process_object("test__", "test__Obj__c", None)
        assert task._should_process_object(
            "test__", "test__Obj__e", metadata_tree.fromstring(object_source_positive)
        )
        assert not task._should_process_object(
            "test__", "test__Obj__c", metadata_tree.fromstring(object_source_negative)
        )
        assert not task._should_process_object(
            "test__", "foo__Obj__c", metadata_tree.fromstring(object_source_positive)
        )
        assert not task._should_process_object(
            "test__", "Account", metadata_tree.fromstring(object_source_positive)
        )

    def test_should_process_object_fields(self):
        object_source_negative = b"""<?xml version="1.0" encoding="UTF-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
    <visibility>Protected</visibility>
    <description>Description</description>
    <label>Test</label>
</CustomObject>"""

        object_source_positive = b"""<?xml version="1.0" encoding="UTF-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
    <description>Description</description>
    <label>Test</label>
</CustomObject>"""

        task = create_task(GenerateDataDictionary, {})

        assert task._should_process_object_fields(
            "test__Obj__c", metadata_tree.fromstring(object_source_positive)
        )
        assert task._should_process_object_fields("test__Obj__c", None)
        assert task._should_process_object_fields(
            "Account", metadata_tree.fromstring(object_source_positive)
        )
        assert task._should_process_object_fields(
            "test__Obj__e", metadata_tree.fromstring(object_source_positive)
        )
        assert not task._should_process_object_fields(
            "test__Obj__c", metadata_tree.fromstring(object_source_negative)
        )

    def test_process_field_element__new(self):
        xml_source = """<?xml version="1.0" encoding="UTF-8"?>
<CustomField xmlns="http://soap.sforce.com/2006/04/metadata">
    <fullName>Lookup__c</fullName>
    <label>Test</label>
    <type>Lookup</type>
    <referenceTo>Test__c</referenceTo>
</CustomField>
"""
        task = create_task(GenerateDataDictionary, {})
        p = Package(
            repo=None, package_name="Test", namespace="test__", prefix_release="rel/"
        )
        v = PackageVersion(package=p, version=LooseVersion("1.1"))

        task._init_schema()
        task._process_field_element(
            "test__Test__c", metadata_tree.fromstring(xml_source.encode("utf-8")), v
        )

        assert "test__Test__c.test__Lookup__c" in task.fields

        assert task.fields["test__Test__c.test__Lookup__c"] == [
            FieldDetail(
                version=v,
                sobject="test__Test__c",
                api_name="test__Lookup__c",
                label="Test",
                type="Lookup to test__Test__c",
                description="",
                help_text="",
                valid_values="",
            )
        ]

    def test_process_field_element__master_detail(self):
        xml_source = """<?xml version="1.0" encoding="UTF-8"?>
<CustomField xmlns="http://soap.sforce.com/2006/04/metadata">
    <fullName>Lookup__c</fullName>
    <label>Test</label>
    <type>MasterDetail</type>
    <referenceTo>Test__c</referenceTo>
</CustomField>
"""
        task = create_task(GenerateDataDictionary, {})
        p = Package(
            repo=None, package_name="Test", namespace="test__", prefix_release="rel/"
        )
        v = PackageVersion(package=p, version=LooseVersion("1.1"))

        task._init_schema()
        task._process_field_element(
            "test__Test__c", metadata_tree.fromstring(xml_source.encode("utf-8")), v
        )

        assert "test__Test__c.test__Lookup__c" in task.fields
        assert task.fields["test__Test__c.test__Lookup__c"] == [
            FieldDetail(
                version=v,
                sobject="test__Test__c",
                api_name="test__Lookup__c",
                label="Test",
                type="Master-Detail Relationship to test__Test__c",
                description="",
                help_text="",
                valid_values="",
            )
        ]

    def test_process_field_element__blank_elements(self):
        xml_source = """<?xml version="1.0" encoding="UTF-8"?>
<CustomField xmlns="http://soap.sforce.com/2006/04/metadata">
    <fullName>Lookup__c</fullName>
    <label>Test</label>
    <description></description>
    <inlineHelpText></inlineHelpText>
    <type>MasterDetail</type>
    <referenceTo>Test__c</referenceTo>
</CustomField>
"""
        task = create_task(GenerateDataDictionary, {})
        p = Package(
            repo=None, package_name="Test", namespace="test__", prefix_release="rel/"
        )
        v = PackageVersion(package=p, version=LooseVersion("1.1"))

        task._init_schema()
        task._process_field_element(
            "test__Test__c", metadata_tree.fromstring(xml_source.encode("utf-8")), v
        )

        assert "test__Test__c.test__Lookup__c" in task.fields
        assert task.fields["test__Test__c.test__Lookup__c"] == [
            FieldDetail(
                version=v,
                sobject="test__Test__c",
                api_name="test__Lookup__c",
                label="Test",
                type="Master-Detail Relationship to test__Test__c",
                description="",
                help_text="",
                valid_values="",
            )
        ]

    def test_process_field_element__standard(self):
        xml_source = """<?xml version="1.0" encoding="UTF-8"?>
<CustomField xmlns="http://soap.sforce.com/2006/04/metadata">
    <fullName>Account</fullName>
    <label>Account</label>
    <type>Lookup</type>
    <referenceTo>Account</referenceTo>
</CustomField>
"""
        task = create_task(GenerateDataDictionary, {})

        task._init_schema()
        p = Package(
            repo=None, package_name="Test", namespace="test__", prefix_release="rel/"
        )
        v = PackageVersion(package=p, version=LooseVersion("1.1"))

        task._process_field_element(
            "test__Test__c", metadata_tree.fromstring(xml_source.encode("utf-8")), v
        )

        assert len(task.fields) == 0

    def test_process_field_element__updated(self):
        xml_source = """<?xml version="1.0" encoding="UTF-8"?>
<CustomField xmlns="http://soap.sforce.com/2006/04/metadata">
    <fullName>Account__c</fullName>
    <inlineHelpText>{}</inlineHelpText>
    <label>Account</label>
    <type>Lookup</type>
    <referenceTo>Account</referenceTo>
</CustomField>
"""
        task = create_task(GenerateDataDictionary, {})

        task._init_schema()
        p = Package(
            repo=None, package_name="Test", namespace="test__", prefix_release="rel/"
        )
        v = PackageVersion(package=p, version=LooseVersion("1.1"))
        v2 = PackageVersion(package=p, version=LooseVersion("1.2"))

        task._process_field_element(
            "test__Test__c",
            metadata_tree.fromstring(xml_source.format("Initial").encode("utf-8")),
            v,
        )

        assert task.fields["test__Test__c.test__Account__c"] == [
            FieldDetail(
                version=v,
                sobject="test__Test__c",
                api_name="test__Account__c",
                label="Account",
                type="Lookup to Account",
                help_text="Initial",
                description="",
                valid_values="",
            )
        ]

        task._process_field_element(
            "test__Test__c",
            metadata_tree.fromstring(xml_source.format("New").encode("utf-8")),
            v2,
        )
        assert task.fields["test__Test__c.test__Account__c"] == [
            FieldDetail(
                version=v,
                sobject="test__Test__c",
                api_name="test__Account__c",
                label="Account",
                type="Lookup to Account",
                help_text="Initial",
                description="",
                valid_values="",
            ),
            FieldDetail(
                version=v2,
                sobject="test__Test__c",
                api_name="test__Account__c",
                label="Account",
                type="Lookup to Account",
                help_text="New",
                description="",
                valid_values="",
            ),
        ]

    def test_process_field_element__valid_values(self):
        xml_source = """<?xml version="1.0" encoding="UTF-8"?>
<CustomField xmlns="http://soap.sforce.com/2006/04/metadata">
    <fullName>Type__c</fullName>
    <label>Type</label>
    <type>Picklist</type>
    <valueSet>
        <valueSetDefinition>
            <value>
                <label>Test 1</label>
                <fullName>Test 1 API</fullName>
            </value>
            <value>
                <label>Test 2</label>
                <fullName>Test 2 API</fullName>
            </value>
        </valueSetDefinition>
    </valueSet>
</CustomField>
"""
        task = create_task(GenerateDataDictionary, {})

        task._init_schema()
        p = Package(
            repo=None, package_name="Test", namespace="test__", prefix_release="rel/"
        )
        v = PackageVersion(package=p, version=LooseVersion("1.1"))

        task._process_field_element(
            "test__Test__c", metadata_tree.fromstring(xml_source.encode("utf-8")), v
        )

        assert task.fields["test__Test__c.test__Type__c"] == [
            FieldDetail(
                version=v,
                sobject="test__Test__c",
                api_name="test__Type__c",
                label="Type",
                type="Picklist",
                help_text="",
                description="",
                valid_values="Test 1; Test 2",
            )
        ]

    def test_process_field_element__valid_values_old_format(self):
        xml_source = """<?xml version="1.0" encoding="UTF-8"?>
<CustomField xmlns="http://soap.sforce.com/2006/04/metadata">
    <fullName>Type__c</fullName>
    <label>Type</label>
    <type>Picklist</type>
    <picklist>
        <picklistValues>
            <fullName>Test 1</fullName>
            <default>false</default>
        </picklistValues>
        <picklistValues>
            <fullName>Test 2</fullName>
            <default>false</default>
        </picklistValues>
    </picklist>
</CustomField>
"""
        task = create_task(GenerateDataDictionary, {})

        task._init_schema()
        p = Package(
            repo=None, package_name="Test", namespace="test__", prefix_release="rel/"
        )
        v = PackageVersion(package=p, version=LooseVersion("1.1"))

        task._process_field_element(
            "test__Test__c", metadata_tree.fromstring(xml_source.encode("utf-8")), v
        )

        assert task.fields["test__Test__c.test__Type__c"] == [
            FieldDetail(
                version=v,
                sobject="test__Test__c",
                api_name="test__Type__c",
                label="Type",
                type="Picklist",
                help_text="",
                description="",
                valid_values="Test 1; Test 2",
            )
        ]

    def test_process_field_element__valid_values_global_value_set(self):
        xml_source = """<?xml version="1.0" encoding="UTF-8"?>
<CustomField xmlns="http://soap.sforce.com/2006/04/metadata">
    <fullName>Type__c</fullName>
    <label>Type</label>
    <type>Picklist</type>
    <valueSet>
        <valueSetName>Test Value Set</valueSetName>
    </valueSet>
</CustomField>
"""
        task = create_task(GenerateDataDictionary, {})

        task._init_schema()
        p = Package(
            repo=None, package_name="Test", namespace="test__", prefix_release="rel/"
        )
        v = PackageVersion(package=p, version=LooseVersion("1.1"))

        task._process_field_element(
            "test__Test__c", metadata_tree.fromstring(xml_source.encode("utf-8")), v
        )

        assert task.fields["test__Test__c.test__Type__c"] == [
            FieldDetail(
                version=v,
                sobject="test__Test__c",
                api_name="test__Type__c",
                label="Type",
                type="Picklist",
                help_text="",
                description="",
                valid_values="Global Value Set Test Value Set",
            )
        ]

    def test_process_field_element__text_length(self):
        xml_source = """<?xml version="1.0" encoding="UTF-8"?>
<CustomField xmlns="http://soap.sforce.com/2006/04/metadata">
    <fullName>Type__c</fullName>
    <label>Type</label>
    <type>Text</type>
    <length>128</length>
</CustomField>
"""
        task = create_task(GenerateDataDictionary, {})

        task._init_schema()
        p = Package(
            repo=None, package_name="Test", namespace="test__", prefix_release="rel/"
        )
        v = PackageVersion(package=p, version=LooseVersion("1.1"))

        task._process_field_element(
            "test__Test__c", metadata_tree.fromstring(xml_source.encode("utf-8")), v
        )

        assert task.fields["test__Test__c.test__Type__c"] == [
            FieldDetail(
                version=v,
                sobject="test__Test__c",
                api_name="test__Type__c",
                label="Type",
                type="Text (128)",
                help_text="",
                description="",
                valid_values="",
            )
        ]

    def test_process_field_element__number_length(self):
        xml_source = """<?xml version="1.0" encoding="UTF-8"?>
<CustomField xmlns="http://soap.sforce.com/2006/04/metadata">
    <fullName>Type__c</fullName>
    <label>Type</label>
    <type>Number</type>
    <precision>18</precision>
    <scale>2</scale>
</CustomField>
"""
        task = create_task(GenerateDataDictionary, {})

        task._init_schema()
        p = Package(
            repo=None, package_name="Test", namespace="test__", prefix_release="rel/"
        )
        v = PackageVersion(package=p, version=LooseVersion("1.1"))

        task._process_field_element(
            "test__Test__c", metadata_tree.fromstring(xml_source.encode("utf-8")), v
        )

        assert task.fields["test__Test__c.test__Type__c"] == [
            FieldDetail(
                version=v,
                sobject="test__Test__c",
                api_name="test__Type__c",
                label="Type",
                type="Number (16.2)",
                help_text="",
                description="",
                valid_values="",
            )
        ]

    def test_process_object_element(self):
        xml_source = """<?xml version="1.0" encoding="UTF-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
    <description>Description</description>
    <label>Test Object</label>
    <fields>
        <fullName>Type__c</fullName>
        <inlineHelpText>Type of field.</inlineHelpText>
        <description>Desc</description>
        <label>Type</label>
        <type>Text</type>
        <length>128</length>
    </fields>
</CustomObject>"""

        task = create_task(GenerateDataDictionary, {})

        task._init_schema()
        p = Package(
            repo=None, package_name="Test", namespace="test__", prefix_release="rel/"
        )
        v = PackageVersion(package=p, version=LooseVersion("1.1"))

        task._process_object_element(
            "test__Test__c", metadata_tree.fromstring(xml_source.encode("utf-8")), v
        )

        assert task.sobjects == {
            "test__Test__c": [
                SObjectDetail(
                    version=v,
                    api_name="test__Test__c",
                    label="Test Object",
                    description="Description",
                )
            ]
        }
        assert task.fields == {
            "test__Test__c.test__Type__c": [
                FieldDetail(
                    version=v,
                    sobject="test__Test__c",
                    api_name="test__Type__c",
                    label="Type",
                    type="Text (128)",
                    help_text="Type of field.",
                    description="Desc",
                    valid_values="",
                )
            ]
        }

    def test_process_object_element__missing_description(self):
        xml_source = """<?xml version="1.0" encoding="UTF-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
    <label>Test Object</label>
    <fields>
        <fullName>Type__c</fullName>
        <inlineHelpText>Type of field.</inlineHelpText>
        <label>Type</label>
        <type>Text</type>
        <length>128</length>
    </fields>
</CustomObject>"""

        task = create_task(GenerateDataDictionary, {})

        task._init_schema()
        p = Package(
            repo=None, package_name="Test", namespace="test__", prefix_release="rel/"
        )
        v = PackageVersion(package=p, version=LooseVersion("1.1"))

        task._process_object_element(
            "test__Test__c", metadata_tree.fromstring(xml_source.encode("utf-8")), v
        )

        assert task.sobjects == {
            "test__Test__c": [
                SObjectDetail(
                    version=v,
                    api_name="test__Test__c",
                    label="Test Object",
                    description="",
                )
            ]
        }

    def test_process_object_element__standard(self):
        xml_source = """<?xml version="1.0" encoding="UTF-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
<description>Description</description>
<label>Test</label>
</CustomObject>"""

        task = create_task(GenerateDataDictionary, {})

        task._init_schema()
        p = Package(
            repo=None, package_name="Test", namespace="test__", prefix_release="rel/"
        )
        v = PackageVersion(package=p, version=LooseVersion("1.1"))

        task._process_object_element(
            "Account", metadata_tree.fromstring(xml_source.encode("utf-8")), v
        )

        assert "Account" not in task.sobjects

    def test_process_object_element__protected_custom_setting(self):
        xml_source = """<?xml version="1.0" encoding="UTF-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
<customSettingsType>List</customSettingsType>
<description>Description</description>
<visibility>Protected</visibility>
<label>Test</label>
</CustomObject>"""

        task = create_task(GenerateDataDictionary, {})

        task._init_schema()
        p = Package(
            repo=None, package_name="Test", namespace="test__", prefix_release="rel/"
        )
        v = PackageVersion(package=p, version=LooseVersion("1.1"))

        task._process_object_element(
            "test__CS__c", metadata_tree.fromstring(xml_source.encode("utf-8")), v
        )

        assert "test__CS__c" not in task.sobjects
        assert task.omit_sobjects == set(["test__CS__c"])

    def test_process_object_element__protected_custom_setting_included(self):
        xml_source = """<?xml version="1.0" encoding="UTF-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
<customSettingsType>List</customSettingsType>
<description>Description</description>
<visibility>Protected</visibility>
<label>Test</label>
</CustomObject>"""

        task = create_task(GenerateDataDictionary, {"include_protected_schema": True})

        task._init_schema()
        p = Package(
            repo=None, package_name="Test", namespace="test__", prefix_release="rel/"
        )
        v = PackageVersion(package=p, version=LooseVersion("1.1"))

        task._process_object_element(
            "test__CS__c", metadata_tree.fromstring(xml_source.encode("utf-8")), v
        )

        assert "test__CS__c" in task.sobjects
        assert "test__CS__c" not in task.omit_sobjects

    def test_process_object_element__protected_custom_setting_old(self):
        xml_source = """<?xml version="1.0" encoding="UTF-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
<customSettingsType>List</customSettingsType>
<description>Description</description>
<customSettingsVisibility>Protected</customSettingsVisibility>
<label>Test</label>
</CustomObject>"""

        task = create_task(GenerateDataDictionary, {})

        task._init_schema()
        p = Package(
            repo=None, package_name="Test", namespace="test__", prefix_release="rel/"
        )
        v = PackageVersion(package=p, version=LooseVersion("1.1"))

        task._process_object_element(
            "test__CS__c", metadata_tree.fromstring(xml_source.encode("utf-8")), v
        )

        assert "test__CS__c" not in task.sobjects
        assert task.omit_sobjects == set(["test__CS__c"])

    def test_process_sfdx_release(self):
        object_source = b"""<?xml version="1.0" encoding="UTF-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
    <description>Description</description>
    <label>Test</label>
</CustomObject>"""
        field_source = b"""<CustomField xmlns="http://soap.sforce.com/2006/04/metadata">
        <fullName>Type__c</fullName>
        <label>Type</label>
        <type>Text</type>
        <length>128</length>
    </CustomField>
"""

        def zip_read(filename):
            if filename.endswith(".object-meta.xml"):
                return object_source

            return field_source

        task = create_task(GenerateDataDictionary, {})

        p = Package(
            repo=None, package_name="Test", namespace="test__", prefix_release="rel/"
        )
        v = PackageVersion(package=p, version=LooseVersion("1.1"))

        zip_file = Mock()
        zip_file.read.side_effect = zip_read
        zip_file.namelist.return_value = [
            "force-app/main/default/objects/Child__c/Child__c.object-meta.xml",
            "force-app/main/default/objects/Child__c/fields/Lookup__c.field-meta.xml",
            "force-app/main/default/objects/Parent__c/Parent__c.object-meta.xml",
            ".gitignore",
            "test__c.object-meta.xml",
        ]
        task._process_object_element = Mock()
        task._process_field_element = Mock()
        task._process_sfdx_release(zip_file, v)

        zip_file.read.assert_has_calls(
            [
                call(
                    "force-app/main/default/objects/Child__c/Child__c.object-meta.xml"
                ),
                call(
                    "force-app/main/default/objects/Child__c/Child__c.object-meta.xml"
                ),
                call(
                    "force-app/main/default/objects/Child__c/fields/Lookup__c.field-meta.xml"
                ),
                call(
                    "force-app/main/default/objects/Parent__c/Parent__c.object-meta.xml"
                ),
            ]
        )

        assert len(task._process_object_element.call_args_list) == 2
        assert task._process_object_element.call_args_list[0][0][0] == "test__Child__c"
        assert task._process_object_element.call_args_list[1][0][0] == "test__Parent__c"

        assert len(task._process_field_element.call_args_list) == 1
        assert task._process_object_element.call_args_list[0][0][0] == "test__Child__c"

    def test_process_sfdx_release__skips_protected_custom_settings_fields(self):
        object_source = b"""<?xml version="1.0" encoding="UTF-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
    <customSettingsType>Hierarchy</customSettingsType>
    <description>Description</description>
    <label>Test</label>
    <visibility>Protected</visibility>
</CustomObject>"""
        field_source = b"""<CustomField xmlns="http://soap.sforce.com/2006/04/metadata">
        <fullName>Type__c</fullName>
        <label>Type</label>
        <type>Text</type>
        <length>128</length>
    </CustomField>
"""

        def zip_read(filename):
            if filename.endswith(".object-meta.xml"):
                return object_source

            return field_source

        task = create_task(GenerateDataDictionary, {})
        task._init_schema()

        p = Package(
            repo=None, package_name="Test", namespace="test__", prefix_release="rel/"
        )
        v = PackageVersion(package=p, version=LooseVersion("1.1"))

        zip_file = Mock()
        zip_file.read.side_effect = zip_read
        zip_file.namelist.return_value = [
            "force-app/main/default/objects/Child__c/Child__c.object-meta.xml",
            "force-app/main/default/objects/Child__c/fields/Lookup__c.field-meta.xml",
            "force-app/main/default/objects/Parent__c/Parent__c.object-meta.xml",
            ".gitignore",
            "test__c.object-meta.xml",
            "sfdx-project.json",
        ]
        task._process_object_element = Mock()
        task._process_field_element = Mock()
        task._process_sfdx_release(zip_file, v)

        task._process_field_element.assert_not_called()

        zip_file.read.assert_has_calls(
            [
                call(
                    "force-app/main/default/objects/Child__c/Child__c.object-meta.xml"
                ),
                call(
                    "force-app/main/default/objects/Child__c/Child__c.object-meta.xml"
                ),
                call(
                    "force-app/main/default/objects/Parent__c/Parent__c.object-meta.xml"
                ),
            ]
        )
        assert task.omit_sobjects == set(["test__Child__c", "test__Parent__c"])

    def test_process_sfdx_release__handles_object_not_found(self):
        field_source = b"""<CustomField xmlns="http://soap.sforce.com/2006/04/metadata">
        <fullName>Type__c</fullName>
        <label>Type</label>
        <type>Text</type>
        <length>128</length>
    </CustomField>
"""

        def zip_read(filename):
            return field_source

        task = create_task(GenerateDataDictionary, {})
        task._init_schema()

        p = Package(
            repo=None, package_name="Test", namespace="test__", prefix_release="rel/"
        )
        v = PackageVersion(package=p, version=LooseVersion("1.1"))

        zip_file = Mock()
        zip_file.read.side_effect = zip_read
        zip_file.namelist.return_value = [
            "force-app/main/default/objects/Child__c/fields/Lookup__c.field-meta.xml",
            "sfdx-project.json",
        ]
        task._process_object_element = Mock()
        task._process_field_element = Mock()
        task._should_process_object_fields = Mock(return_value=True)

        task._process_sfdx_release(zip_file, v)

        task._should_process_object_fields.assert_called_once_with(
            "test__Child__c", None
        )

    @patch("cumulusci.tasks.datadictionary.download_extract_vcs_from_repo")
    def test_walk_releases__mdapi(self, extract_github):
        project_config = create_project_config()
        project_config.project__git__prefix_release = "rel/"
        project_config.project__name = "Project"
        task = create_task(GenerateDataDictionary, {}, project_config=project_config)
        task._init_schema()

        repo = Mock()
        release = Mock()
        release.draft = False
        release.prerelease = False
        release.tag_name = "rel/1.1"
        repo.releases.return_value = [release]
        task._process_mdapi_release = Mock()
        extract_github.return_value.namelist.return_value = ["src/objects/"]
        p = Package(
            repo=repo, package_name="Test", namespace="test__", prefix_release="rel/"
        )

        task._walk_releases(p)

        task._process_mdapi_release.assert_called_once_with(
            extract_github.return_value,
            PackageVersion(package=p, version=LooseVersion("1.1")),
        )

    @patch("cumulusci.tasks.datadictionary.download_extract_vcs_from_repo")
    def test_walk_releases__sfdx(self, extract_github):
        project_config = create_project_config()
        project_config.project__git__prefix_release = "rel/"
        project_config.project__name = "Project"

        task = create_task(GenerateDataDictionary, {}, project_config=project_config)
        task._init_schema()

        repo = Mock()
        release = Mock()
        release.draft = False
        release.prerelease = False
        release.tag_name = "rel/1.1"
        repo.releases.return_value = [release]
        task._process_sfdx_release = Mock()
        extract_github.return_value.namelist.return_value = [
            "force-app/main/default/objects/",
            "sfdx-project.json",
        ]
        p = Package(
            repo=repo, package_name="Test", namespace="test__", prefix_release="rel/"
        )

        task._walk_releases(p)

        task._process_sfdx_release.assert_called_once_with(
            extract_github.return_value,
            PackageVersion(package=p, version=LooseVersion("1.1")),
        )

    @patch("cumulusci.tasks.datadictionary.download_extract_vcs_from_repo")
    def test_walk_releases__draft(self, extract_github):
        project_config = create_project_config()
        project_config.project__git__prefix_release = "rel/"
        project_config.project__name = "Project"
        task = create_task(GenerateDataDictionary, {}, project_config=project_config)
        task._init_schema()

        repo = Mock()
        release_draft = Mock()
        release_draft.draft = False
        release_draft.prerelease = False
        release_draft.tag_name = "uat/1.1_Beta_1"
        release_real = Mock()
        release_real.draft = False
        release_real.prerelease = False
        release_real.tag_name = "rel/1.1"

        repo.releases.return_value = [release_draft, release_real]
        task._process_zipfile = Mock()
        p = Package(
            repo=repo, package_name="Test", namespace="test__", prefix_release="rel/"
        )

        task._walk_releases(p)

        task._process_zipfile.assert_called_once()

    @patch("cumulusci.tasks.datadictionary.download_extract_vcs_from_repo")
    def test_walk_releases__prerelease(self, extract_github):
        project_config = create_project_config()
        project_config.project__git__prefix_release = "rel/"
        project_config.project__name = "Project"
        project_config.repo_info["branch"] = "feature/foo"
        task = create_task(
            GenerateDataDictionary,
            {"include_prerelease": True},
            project_config=project_config,
        )
        task._init_schema()

        repo = Mock()
        release = Mock()
        release.draft = False
        release.prerelease = False
        release.tag_name = "rel/1.1"
        repo.releases.return_value = [release]
        task._process_mdapi_release = Mock()
        extract_github.return_value.namelist.return_value = ["src/objects/"]
        p = Package(
            repo=repo, package_name="Test", namespace="test__", prefix_release="rel/"
        )

        task._walk_releases(p)

        extract_github.assert_has_calls(
            [call(repo, ref="rel/1.1"), call(repo, ref="feature/foo")], any_order=True
        )
        task._process_mdapi_release.assert_has_calls(
            [
                call(
                    extract_github.return_value,
                    PackageVersion(package=p, version=LooseVersion("1.1")),
                ),
                call(
                    extract_github.return_value,
                    PackageVersion(package=p, version=PRERELEASE_SIGIL),
                ),
            ]
        )

    def test_init_schema(self):
        task = create_task(GenerateDataDictionary, {})
        task._init_schema()

        assert task.fields is not None
        assert task.sobjects is not None

    def test_run_task__additional_dependencies(self):
        project_config = create_project_config()
        project_config.keychain.get_service = Mock()
        project_config.project__package__name = "Project"
        project_config.project__name = "Project"
        project_config.project__package__namespace = "test"
        project_config.project__dependencies = [
            {"github": "https://github.com/test/test"}
        ]

        task = create_task(
            GenerateDataDictionary,
            {"additional_dependencies": [{"github": "https://github.com/test/test"}]},
            project_config=project_config,
        )
        task.get_repo = Mock()
        release = Mock()
        release.draft = False
        release.prerelease = False
        release.tag_name = "release/1.1"
        task.get_repo.return_value.releases.return_value = [release]
        task._get_repo_dependencies = Mock(return_value=[1, 2])
        task._walk_releases = Mock()

        with temporary_dir():
            task._run_task()

        task._get_repo_dependencies.assert_has_calls(
            [
                call(
                    [
                        GitHubDynamicDependency(github="https://github.com/test/test"),
                        GitHubDynamicDependency(github="https://github.com/test/test"),
                    ]
                ),
            ]
        )

        task._walk_releases.assert_has_calls(
            [
                call(
                    Package(
                        repo=task.get_repo.return_value,
                        package_name=project_config.project__package__name,
                        namespace="test__",
                        prefix_release="release/",
                    )
                ),
                call(1),
                call(2),
            ]
        )

    @patch("cumulusci.tasks.datadictionary.download_extract_vcs_from_repo")
    def test_run_task(self, extract_github):
        # This is an integration test. We mock out `get_repo()` and the filesystem.
        xml_source = """<?xml version="1.0" encoding="UTF-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
    <description>Description</description>
    <label>Test</label>
    <fields>
        <fullName>Type__c</fullName>
        <inlineHelpText>Type of field.</inlineHelpText>
        <label>Type</label>
        <type>Text</type>
        <length>255</length>
    </fields>
</CustomObject>"""
        project_config = create_project_config()
        project_config.keychain.get_service = Mock()
        project_config.project__package__name = "Project"
        project_config.project__name = "Project"
        project_config.project__package__namespace = "test"

        task = create_task(GenerateDataDictionary, project_config=project_config)

        task.get_repo = Mock()
        release = Mock()
        release.draft = False
        release.prerelease = False
        release.tag_name = "release/1.1"
        task.get_repo.return_value.releases.return_value = [release]

        extract_github.return_value.namelist.return_value = [
            "src/objects/",
            "src/objects/Test__c.object",
        ]
        extract_github.return_value.read.return_value = xml_source.encode("utf-8")
        m = mock_open()

        with patch("builtins.open", m):
            task()

        m.assert_has_calls(
            [call("Project Objects.csv", "w"), call("Project Fields.csv", "w")],
            any_order=True,
        )

        m.return_value.write.assert_has_calls(
            [
                call(
                    '"Object Label","Object API Name","Object Description","Version Introduced","Version Deleted"\r\n'
                ),
                call('"Test","test__Test__c","Description","Project 1.1",""\r\n'),
                call(
                    '"Object Label","Object API Name","Field Label","Field API Name","Type","Picklist Values","Help Text","Field Description","Version Introduced","Version Picklist Values Last Changed","Version Help Text Last Changed","Version Deleted"\r\n'
                ),
                call(
                    '"Test","test__Test__c","Type","test__Type__c","Text (255)","","Type of field.","","Project 1.1","","",""\r\n'
                ),
            ],
            any_order=True,
        )

    @patch("cumulusci.tasks.datadictionary.download_extract_vcs_from_repo")
    def test_run_task__prerelease(self, extract_github):
        # This is an integration test. We mock out `get_repo()` and the filesystem.
        xml_source = """<?xml version="1.0" encoding="UTF-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
    <description>Description</description>
    <label>Test</label>
    <fields>
        <fullName>Type__c</fullName>
        <inlineHelpText>Type of field.</inlineHelpText>
        <label>Type</label>
        <type>Text</type>
        <length>255</length>
    </fields>
</CustomObject>"""
        xml_source_prerelease = """<?xml version="1.0" encoding="UTF-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
    <description>Description</description>
    <label>Test</label>
    <fields>
        <fullName>Type__c</fullName>
        <inlineHelpText>Type of field.</inlineHelpText>
        <label>Type</label>
        <type>Text</type>
        <length>255</length>
    </fields>
    <fields>
        <fullName>Description__c</fullName>
        <inlineHelpText>Description of field.</inlineHelpText>
        <label>Description</label>
        <type>Text</type>
        <length>255</length>
    </fields>
</CustomObject>"""

        project_config = create_project_config()
        project_config.keychain.get_service = Mock()
        project_config.project__package__name = "Project"
        project_config.project__name = "Project"
        project_config.project__package__namespace = "test"
        project_config.repo_info["branch"] = "testbranch"

        task = create_task(
            GenerateDataDictionary,
            {"include_prerelease": True},
            project_config=project_config,
        )

        task.get_repo = Mock()
        release = Mock()
        release.draft = False
        release.prerelease = False
        release.tag_name = "release/1.1"
        task.get_repo.return_value.releases.return_value = [release]

        extract_github.return_value.namelist.return_value = [
            "src/objects/",
            "src/objects/Test__c.object",
        ]
        extract_github.return_value.read.side_effect = [
            xml_source.encode("utf-8"),
            xml_source_prerelease.encode("utf-8"),
        ]
        m = mock_open()

        with patch("builtins.open", m):
            task()

        m.assert_has_calls(
            [call("Project Objects.csv", "w"), call("Project Fields.csv", "w")],
            any_order=True,
        )

        m.return_value.write.assert_has_calls(
            [
                call(
                    '"Object Label","Object API Name","Object Description","Version Introduced","Version Deleted"\r\n'
                ),
                call('"Test","test__Test__c","Description","Project 1.1",""\r\n'),
                call(
                    '"Object Label","Object API Name","Field Label","Field API Name","Type","Picklist Values","Help Text","Field Description","Version Introduced","Version Picklist Values Last Changed","Version Help Text Last Changed","Version Deleted"\r\n'
                ),
                call(
                    '"Test","test__Test__c","Type","test__Type__c","Text (255)","","Type of field.","","Project 1.1","","",""\r\n'
                ),
                call(
                    '"Test","test__Test__c","Description","test__Description__c","Text (255)","","Description of field.","","Project Prerelease","","",""\r\n'
                ),
            ],
            any_order=True,
        )

    def test_init_options(self):
        task = create_task(
            GenerateDataDictionary,
            {"object_path": "objects.csv", "field_path": "fields.csv"},
        )

        assert task.options["object_path"] == "objects.csv"
        assert task.options["field_path"] == "fields.csv"

    def test_init_options__defaults(self):
        project_config = create_project_config()
        project_config.project__name = "Project"

        task = create_task(GenerateDataDictionary, {}, project_config)

        assert task.options["object_path"] == "Project Objects.csv"
        assert task.options["field_path"] == "Project Fields.csv"

    def test_init_options__bad_deps(self):
        project_config = create_project_config()
        project_config.project__name = "Project"

        with pytest.raises(DependencyParseError):
            create_task(
                GenerateDataDictionary,
                {"additional_dependencies": [{"namespace": "foo"}]},
                project_config,
            )

    def test_init_options__non_github_deps(self):
        project_config = create_project_config()
        project_config.project__name = "Project"

        with pytest.raises(TaskOptionsError):
            create_task(
                GenerateDataDictionary,
                {"additional_dependencies": [{"namespace": "foo", "version": "1.0"}]},
                project_config,
            )

    def test_init_options__prerelease(self):
        project_config = create_project_config()
        project_config.project__name = "Project"

        with pytest.raises(TaskOptionsError):
            create_task(
                GenerateDataDictionary,
                {
                    "include_prerelease": True,
                    "additional_dependencies": [
                        {"github": "http://github.com/test/test"}
                    ],
                },
                project_config,
            )

        task = create_task(
            GenerateDataDictionary,
            {"include_prerelease": True, "include_dependencies": True},
            project_config,
        )

        assert task.options["include_dependencies"] is False

    def test_get_repo_dependencies__none(self):
        task = create_task(
            GenerateDataDictionary,
            {
                "object_path": "object.csv",
                "field_path": "fields.csv",
                "release_prefix": "rel/",
            },
        )

        assert task._get_repo_dependencies([]) == []

    @patch("cumulusci.vcs.github.service.GitHubService")
    @patch("cumulusci.vcs.github.service.GitHubEnterpriseService")
    def test_get_repo_dependencies__cannot_get_repo(
        self, mock_github_service, mock_github_enterprise_service
    ):
        mock_github_service.get_service_for_url.return_value = None
        mock_github_enterprise_service.get_service_for_url.return_value = None

        project_config = create_project_config()
        project_config.project__git__prefix_release = "rel/"
        project_config.project__name = "Project"

        task = create_task(
            GenerateDataDictionary,
            {
                "object_path": "object.csv",
                "field_path": "fields.csv",
                "release_prefix": "rel/",
            },
            project_config=project_config,
        )

        project_config.project__dependencies = [
            {"github": "https://github.com/test/test"}
        ]
        project_config.get_repo_from_url = Mock(return_value=None)

        with pytest.raises(DependencyResolutionError):
            task._get_repo_dependencies(
                parse_dependencies(project_config.project__dependencies)
            )

    @patch("cumulusci.tasks.datadictionary.get_static_dependencies")
    @patch("cumulusci.tasks.datadictionary.get_repo_from_url")
    @patch("cumulusci.tasks.datadictionary.get_remote_project_config")
    def test_get_repo_dependencies__success(
        self, get_remote_project_config, get_repo_from_url, get_static_dependencies
    ):
        project_config = create_project_config()
        project_config.project__git__prefix_release = "rel/"
        project_config.project__name = "Project"

        task = create_task(GenerateDataDictionary, {}, project_config=project_config)

        project_config.project__dependencies = [
            {"github": "https://github.com/test/test"}
        ]

        cumulusci_yml_one = io.StringIO(
            """
project:
    name: Test 1
    package:
        name: Test 1
        namespace: test1
    dependencies:
        - github: "https://github.com/test1/test1"
"""
        )
        cumulusci_yml_two = io.StringIO(
            """
project:
    name: Test 2
    package:
        name: Test 2
    dependencies:
        - github: "test1"
    git:
        prefix_release: "rel/"
"""
        )

        def fake_get_static_dependencies(
            context,
            dependencies=None,
            resolution_strategy=None,
            strategies=None,
            filter_function=None,
        ):
            filter_function(
                GitHubDynamicDependency(github="https://github.com/test/test")
            ),
            filter_function(
                GitHubDynamicDependency(github="https://github.com/test1/test1")
            )
            return [
                GitHubDynamicDependency(github="https://github.com/test/test"),
                GitHubDynamicDependency(github="https://github.com/test1/test1"),
            ]

        get_static_dependencies.side_effect = fake_get_static_dependencies

        get_remote_project_config.side_effect = [
            BaseProjectConfig(None, cci_safe_load(cumulusci_yml_one)),
            BaseProjectConfig(None, cci_safe_load(cumulusci_yml_two)),
        ]

        results = task._get_repo_dependencies(
            parse_dependencies(project_config.project__dependencies)
        )

        assert results == [
            Package(
                repo=get_repo_from_url.return_value,
                package_name="Test 1",
                namespace="test1__",
                prefix_release="release/",
            ),
            Package(
                repo=get_repo_from_url.return_value,
                package_name="Test 2",
                namespace="",
                prefix_release="rel/",
            ),
        ]



import os
from typing import Any, Dict, List, Optional

from pydantic.v1 import root_validator

from cumulusci.core.exceptions import SalesforceDXException
from cumulusci.salesforce_api.utils import get_simple_salesforce_connection
from cumulusci.tasks.salesforce import BaseSalesforceApiTask
from cumulusci.utils.options import CCIOptions, Field


class ExtParameter(CCIOptions):
    """External Auth Identity Provider Parameter options"""

    name: str = Field(
        None,
        description="Parameter name. [default to None]",
    )
    value: str = Field(
        None,
        description="Parameter value. [default to None]",
    )
    sequence_number: int = Field(
        None,
        description="Sequence number. [default to None]",
    )


class ExternalAuthIdentityProviderCredential(CCIOptions):
    """External Auth Identity Provider Credential options"""

    name: str = Field(
        None,
        description="Credential name. [default to None]",
    )
    client_id: str = Field(
        None,
        description="Client ID. [default to None]",
    )
    client_secret: str = Field(
        None,
        description="Client secret. [default to None]",
    )
    auth_protocol: str = Field(
        "OAuth",
        description="Authentication protocol. [default to OAuth]",
    )


class ExternalAuthIdentityProviderParameter(CCIOptions):
    """External Auth Identity Provider Parameter options"""

    authorize_url: str = Field(
        None,
        description="Authorize URL. [default to None]",
    )
    token_url: str = Field(
        None,
        description="Token URL. [default to None]",
    )
    user_info_url: str = Field(
        None,
        description="User info URL. [default to None]",
    )
    jwks_url: str = Field(
        None,
        description="JWKS URL (for OpenID Connect). [default to None]",
    )
    issuer_url: str = Field(
        None,
        description="Issuer URL (for OpenID Connect). [default to None]",
    )
    client_authentication: str = Field(
        None,
        description="Client authentication method (e.g., ClientSecretBasic, ClientSecretPost, ClientSecretJwt). [default to None]",
    )
    custom_parameter: ExtParameter = Field(
        None,
        description="Custom parameter. [default to None]",
    )
    identity_provider_option: ExtParameter = Field(
        None,
        description="Identity provider option (e.g., PkceEnabled, UserinfoEnabled). [default to None]",
    )
    credential: ExternalAuthIdentityProviderCredential = Field(
        None,
        description="Credential to update. [default to None]",
    )
    secret: bool = Field(
        False,
        description="Is the value a secret. [default to False]",
    )

    @root_validator
    def check_parameters(cls, values):
        """Check if at least one parameter is provided"""
        param_fields = [
            "authorize_url",
            "token_url",
            "user_info_url",
            "jwks_url",
            "issuer_url",
            "client_authentication",
            "custom_parameter",
            "identity_provider_option",
            "credential",
        ]

        provided_params = [
            field for field in param_fields if values.get(field) is not None
        ]

        if len(provided_params) == 0:
            raise ValueError("At least and only one parameter must be provided.")

        if len(provided_params) > 1:
            raise ValueError("At least and only one parameter must be provided.")

        return values

    def get_external_auth_identity_provider_parameter(self):
        """Get the external auth identity provider parameter based on which field is set"""
        ext_auth_param = {}

        if self.authorize_url is not None:
            ext_auth_param["parameterType"] = "AuthorizeUrl"
            ext_auth_param["parameterName"] = "AuthorizeUrl"
            ext_auth_param["parameterValue"] = self.authorize_url

        if self.token_url is not None:
            ext_auth_param["parameterType"] = "TokenUrl"
            ext_auth_param["parameterName"] = "TokenUrl"
            ext_auth_param["parameterValue"] = self.token_url

        if self.user_info_url is not None:
            ext_auth_param["parameterType"] = "UserInfoUrl"
            ext_auth_param["parameterName"] = "UserInfoUrl"
            ext_auth_param["parameterValue"] = self.user_info_url

        if self.jwks_url is not None:
            ext_auth_param["parameterType"] = "JwksUrl"
            ext_auth_param["parameterName"] = "JwksUrl"
            ext_auth_param["parameterValue"] = self.jwks_url

        if self.issuer_url is not None:
            ext_auth_param["parameterType"] = "IssuerUrl"
            ext_auth_param["parameterName"] = "IssuerUrl"
            ext_auth_param["parameterValue"] = self.issuer_url

        if self.client_authentication is not None:
            ext_auth_param["parameterType"] = "ClientAuthentication"
            ext_auth_param["parameterName"] = "ClientAuthentication"
            ext_auth_param["parameterValue"] = self.client_authentication

        if self.custom_parameter is not None:
            ext_auth_param["parameterType"] = "CustomParameter"
            ext_auth_param["parameterName"] = self.custom_parameter.name
            ext_auth_param["parameterValue"] = self.custom_parameter.value
            if self.custom_parameter.sequence_number is not None:
                ext_auth_param["sequenceNumber"] = self.custom_parameter.sequence_number

        if self.identity_provider_option is not None:
            ext_auth_param["parameterType"] = "IdentityProviderOptions"
            ext_auth_param["parameterName"] = self.identity_provider_option.name
            ext_auth_param["parameterValue"] = self.identity_provider_option.value

        return ext_auth_param

    def get_credential(self, ext_auth_identity_provider_full_name: str):
        """Get the credential to update"""
        if self.credential is None:
            return None

        return {
            "credentials": [
                {
                    "credentialName": "clientId",
                    "credentialValue": self.credential.client_id,
                },
                {
                    "credentialName": "clientSecret",
                    "credentialValue": self.credential.client_secret,
                },
            ]
        }


class TransformExternalAuthIdentityProviderParameter(
    ExternalAuthIdentityProviderParameter
):
    """Transform External Auth Identity Provider Parameter with environment variable support"""

    def get_external_auth_identity_provider_parameter(self):
        ret = super().get_external_auth_identity_provider_parameter()
        if ret.get("parameterValue", None) is not None:
            ret["parameterValue"] = os.getenv(
                ret.get("parameterValue"), ret.get("parameterValue")
            )
        return ret

    def get_credential(self, ext_auth_identity_provider_full_name: str):
        """Get the credential with values from environment variables"""
        value = super().get_credential(ext_auth_identity_provider_full_name)

        if value is None:
            return None

        for credential in value["credentials"]:
            if credential["credentialValue"]:
                credential["credentialValue"] = os.getenv(
                    credential["credentialValue"], credential["credentialValue"]
                )

        return value


ExternalAuthIdentityProviderParameter.update_forward_refs()
TransformExternalAuthIdentityProviderParameter.update_forward_refs()


class UpdateExternalAuthIdentityProvider(BaseSalesforceApiTask):
    """Custom task to update external auth identity provider parameters.
    This task updates External Auth Identity Provider parameters and credentials using
    the Tooling API and Connect API.

    Reference:
    - https://developer.salesforce.com/docs/atlas.en-us.api_tooling.meta/api_tooling/tooling_api_objects_externalauthidentityprovider.htm
    - https://developer.salesforce.com/docs/atlas.en-us.chatterapi.meta/chatterapi/connect_resources_named_credentials_external_auth_identity_provider_credentials.htm
    """

    class Options(CCIOptions):
        name: str = Field(
            ..., description="Name of the external auth identity provider to update."
        )
        namespace: str = Field(
            "",
            description="Namespace of the external auth identity provider to update. [default to empty string]",
        )
        # External auth identity provider parameters
        parameters: List[ExternalAuthIdentityProviderParameter] = Field(
            [],
            description="Parameters to update. [default to empty list]",
        )
        # Transform parameters (from environment variables)
        transform_parameters: List[
            TransformExternalAuthIdentityProviderParameter
        ] = Field(
            [],
            description="Parameters to transform from environment variables. [default to empty list]",
        )

    parsed_options: Options

    def _init_task(self):
        self.tooling = get_simple_salesforce_connection(
            self.project_config,
            self.org_config,
            api_version=self.project_config.project__package__api_version,
            base_url="tooling",
        )
        self.connect = get_simple_salesforce_connection(
            self.project_config,
            self.org_config,
            api_version=self.project_config.project__package__api_version,
        )

    def _run_task(self):
        # Step 1: Get the external auth identity provider id from the name
        ext_auth_id = self._get_external_auth_identity_provider_id()

        if not ext_auth_id:
            msg = f"External auth identity provider '{self.parsed_options.name}' not found"
            raise SalesforceDXException(msg)

        # Step 2: Get the external auth identity provider object
        ext_auth_provider = self._get_external_auth_identity_provider_object(
            ext_auth_id
        )

        if not ext_auth_provider:
            msg = f"Failed to retrieve external auth identity provider object for '{self.parsed_options.name}'"
            raise SalesforceDXException(msg)

        # Step 3: Update the external auth identity provider parameters
        self._update_external_auth_identity_provider_parameters(ext_auth_provider)

        updated_ext_auth_provider = self._update_external_auth_identity_provider_object(
            ext_auth_id, ext_auth_provider
        )

        if not updated_ext_auth_provider:
            msg = f"Failed to update external auth identity provider object for '{self.parsed_options.name}'"
            raise SalesforceDXException(msg)

        # Step 4: Update credentials if specified
        response = self._update_credential()

        if not response:
            raise SalesforceDXException(
                f"Failed to update credentials for external auth identity provider '{self.parsed_options.name}'"
            )

        self.logger.info(
            f"Successfully updated external auth identity provider '{self.parsed_options.name}'"
        )

    def _get_external_auth_identity_provider_id(self) -> Optional[str]:
        """Get the external auth identity provider ID from the name"""
        query = f"SELECT Id FROM ExternalAuthIdentityProvider WHERE DeveloperName='{self.parsed_options.name}'"

        if self.parsed_options.namespace:
            query += f" AND NamespacePrefix='{self.parsed_options.namespace}'"

        query += " LIMIT 1"

        try:
            res = self.tooling.query(query)
            if res["size"] == 0:
                return None
            return res["records"][0]["Id"]
        except Exception as e:
            self.logger.error(
                f"Error querying external auth identity provider: {str(e)}"
            )
            return None

    def _get_external_auth_identity_provider_object(
        self, ext_auth_id: str
    ) -> Optional[Dict[str, Any]]:
        """Get the external auth identity provider object using Tooling API"""
        try:
            # Use Tooling API to get the external auth identity provider metadata
            result = self.tooling._call_salesforce(
                method="GET",
                url=f"{self.tooling.base_url}sobjects/ExternalAuthIdentityProvider/{ext_auth_id}",
            )

            if result.status_code != 200:
                self.logger.error(
                    f"Error retrieving external auth identity provider object: {result.json()}"
                )
                return None

            return result.json().get("Metadata", None)

        except Exception as e:
            self.logger.error(
                f"Error retrieving external auth identity provider object: {str(e)}"
            )
            return None

    def _update_external_auth_identity_provider_object(
        self, ext_auth_id: str, ext_auth_provider: Dict[str, Any]
    ) -> Optional[bool]:
        """Update the external auth identity provider object"""
        try:
            result_update = self.tooling._call_salesforce(
                method="PATCH",
                url=f"{self.tooling.base_url}sobjects/ExternalAuthIdentityProvider/{ext_auth_id}",
                json={"Metadata": ext_auth_provider},
            )
            if not result_update.ok:
                self.logger.error(
                    f"Error updating external auth identity provider object: {result_update.json()}"
                )
                return None

            return result_update.ok
        except Exception as e:
            self.logger.error(
                f"Error updating external auth identity provider object: {str(e)}"
            )
            return None

    def _update_external_auth_identity_provider_parameters(
        self, ext_auth_provider: Dict[str, Any]
    ):
        """Update the external auth identity provider parameters"""
        try:
            # Get template parameter for new parameters
            template_param = (
                self._get_external_auth_identity_provider_template_parameter()
            )

            # Update regular parameters
            self._update_parameters(
                ext_auth_provider, self.parsed_options.parameters, template_param
            )

            # Update transform parameters (from environment variables)
            self._update_parameters(
                ext_auth_provider,
                self.parsed_options.transform_parameters,
                template_param,
            )

        except Exception as e:
            raise SalesforceDXException(f"Failed to update parameters: {str(e)}")

    def _update_parameters(
        self,
        ext_auth_provider: Dict[str, Any],
        ext_auth_parameters: List[ExternalAuthIdentityProviderParameter],
        template_param: Dict[str, Any],
    ):
        """Update the parameters"""
        for param_input in ext_auth_parameters:
            # Skip credential-only updates
            if param_input.credential is not None and all(
                getattr(param_input, field) is None
                for field in [
                    "authorize_url",
                    "token_url",
                    "user_info_url",
                    "jwks_url",
                    "issuer_url",
                    "client_authentication",
                    "custom_parameter",
                    "identity_provider_option",
                ]
            ):
                continue

            param_to_update = (
                param_input.get_external_auth_identity_provider_parameter()
            )
            secret = (
                param_to_update.pop("secret", False)
                if "secret" in param_to_update
                else False
            )

            # Create a copy for matching (without parameterValue and parameterName)
            param_to_match = {
                k: v
                for k, v in param_to_update.items()
                if k != "parameterValue" and k != "parameterName"
            }

            # Find existing parameter
            auth_param = next(
                (
                    param
                    for param in ext_auth_provider.get(
                        "externalAuthIdentityProviderParameters", []
                    )
                    if param_to_match.items() <= param.items()
                ),
                None,
            )

            if auth_param:
                # Update existing parameter
                auth_param.update(param_to_update)
                self.logger.info(
                    f"Updated parameter {auth_param['parameterType']}"
                    + (
                        f"-{auth_param.get('parameterName', 'N/A')}"
                        if auth_param.get("parameterName")
                        else ""
                    )
                    + f" with new value {param_to_update['parameterValue'] if not secret else '********'}"
                )
            else:
                # Add new parameter
                copy_template_param = template_param.copy()
                copy_template_param.update(param_to_update)
                if "externalAuthIdentityProviderParameters" not in ext_auth_provider:
                    ext_auth_provider["externalAuthIdentityProviderParameters"] = []
                ext_auth_provider["externalAuthIdentityProviderParameters"].append(
                    copy_template_param
                )
                self.logger.info(
                    f"Added parameter {copy_template_param['parameterType']}"
                    + (
                        f"-{copy_template_param.get('parameterName', 'N/A')}"
                        if copy_template_param.get("parameterName")
                        else ""
                    )
                    + f" with new value {param_to_update['parameterValue'] if not secret else '********'}"
                )

    def _get_external_auth_identity_provider_template_parameter(
        self,
    ) -> Dict[str, Any]:
        """Get the external auth identity provider template parameter"""
        return {
            "description": None,
            "parameterName": None,
            "parameterType": None,
            "parameterValue": None,
            "sequenceNumber": None,
        }

    def _update_credential(self):
        """Update the credential using Connect API"""
        for param in (
            self.parsed_options.parameters + self.parsed_options.transform_parameters
        ):
            if param.credential is None or (
                param.credential.client_secret is None
                and param.credential.client_id is None
            ):
                continue

            namespace = (
                f"{self.parsed_options.namespace}__"
                if self.parsed_options.namespace
                else ""
            )
            ext_auth_full_name = f"{namespace}{self.parsed_options.name}"

            self.logger.info(
                f"Managing credential for external auth identity provider {self.parsed_options.name}..."
            )

            # Get current credential
            credential_response = self.connect._call_salesforce(
                method="GET",
                url=f"{self.connect.base_url}named-credentials/external-auth-identity-provider-credentials/{ext_auth_full_name}",
            )

            if not credential_response.ok:
                msg = f"Failed to retrieve credential for {self.parsed_options.name}: {credential_response.json()}"
                raise SalesforceDXException(msg)

            credential = credential_response.json()
            http_verb = "PUT" if credential.get("credentials") else "POST"

            # Update credential data
            credential_data = param.get_credential(ext_auth_full_name)
            credential["credentials"] = credential_data["credentials"]

            # Update credential via Connect API
            try:
                response = self.connect._call_salesforce(
                    method=http_verb,
                    url=f"{self.connect.base_url}named-credentials/external-auth-identity-provider-credentials/{ext_auth_full_name}",
                    json=credential,
                )

                return response.ok

            except Exception as e:
                self.logger.error(
                    f"Error updating credential for {self.parsed_options.name}: {str(e)}"
                )
                return False

        # Return True if no credentials to update or all updates succeeded
        return True



import typing as T
import xml.etree.ElementTree as etree
from pathlib import Path

from lxml import etree as lxml_etree

UTF8 = "UTF-8"


def elementtree_parse_file(
    path: T.Union[str, Path, T.IO], namespace=None
) -> etree.ElementTree:
    """Parse a file from filename, Path or Stream using Python stdlib.

    All else equal, prefer elementtree over LXML for performance and simplicity reasons."""
    try:
        if namespace:
            etree.register_namespace("", namespace)
        tree = etree.parse(path)
    except etree.ParseError as err:
        err.filename = path
        raise err
    return tree


def lxml_parse_file(path: T.Union[str, Path, T.IO]) -> lxml_etree._ElementTree:
    """Parse a file from filename, Path or stream using lxml for richer API

    Use this if you need advanced xpath and parent-pointer features.
    Otherwise prefer elementree_parse_file for performance and simplicity reasons."""
    parser = lxml_etree.XMLParser(
        resolve_entities=False, load_dtd=False, no_network=True
    )
    try:
        if isinstance(path, Path):
            path = str(path)
        tree = lxml_etree.parse(path, parser=parser)
    except etree.ParseError as err:
        err.filename = path
        raise err
    return tree


elementtree_parse_string = etree.fromstring


def lxml_parse_string(string: str) -> lxml_etree._ElementTree:
    """Parse a string using lxml for richer API

    Use this if you need advanced xpath and parent-pointer features.
    Otherwise prefer elementree_parse_string for performance and simplicity reasons."""

    parser = lxml_etree.XMLParser(
        resolve_entities=False, load_dtd=False, no_network=True
    )
    return lxml_etree.ElementTree(lxml_etree.fromstring(string, parser=parser))


def remove_xml_element_file(name: str, path: T.Union[str, Path]):
    """Remove XML elements from a single file"""
    etree.register_namespace("", "http://soap.sforce.com/2006/04/metadata")
    tree = elementtree_parse_file(path)
    tree = remove_xml_element(name, tree)
    return tree.write(path, encoding=UTF8, xml_declaration=True)


def remove_xml_element_string(name: str, content: str):
    """Remove XML elements from a string"""
    etree.register_namespace("", "http://soap.sforce.com/2006/04/metadata")
    tree = etree.fromstring(content)
    tree = remove_xml_element(name, tree)
    clean_content = etree.tostring(tree, encoding=UTF8)
    return clean_content


def remove_xml_element(name: str, tree):
    """Removes XML elements from an ElementTree content tree"""
    # root = tree.getroot()
    remove = tree.findall(
        ".//{{http://soap.sforce.com/2006/04/metadata}}{}".format(name)
    )
    if not remove:
        return tree

    parent_map = {c: p for p in tree.iter() for c in p}

    for elem in remove:
        parent = parent_map[elem]
        parent.remove(elem)

    return tree
